#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Compatibility layer for Qiskit."""

from typing import Any

from qiskit.quantum_info import Pauli, SparsePauliOp

from . import _fast_pauli as fp


def from_qiskit(qiskit_type: Any) -> Any:
    """Convert Qiskit type to fast_pauli type.

    Parameters
    ----------
    qiskit_type : Any
        Qiskit type to convert to fast_pauli type.

    Returns
    -------
    Any
        fast_pauli type.

    Raises
    ------
    NotImplementedError
        Conversion from Qiskit type to fast_pauli type isn't supported.
    """
    if isinstance(qiskit_type, SparsePauliOp):
        tuples = qiskit_type.to_list(array=True)
        coeffs = [c for _, c in tuples]
        strings = [s for s, _ in tuples]
        return fp.PauliOp(coeffs, strings)

    else:
        raise NotImplementedError(
            f"Conversion from {type(qiskit_type)} to "
            + "fast_pauli equivalent isn't supported."
        )


def to_qiskit(fast_pauli_type: Any) -> Any:
    """Convert fast_pauli type to Qiskit type.

    Parameters
    ----------
    fast_pauli_type : Any
        fast_pauli type to convert to Qiskit type.

    Returns
    -------
    Any
        Qiskit type.

    Raises
    ------
    NotImplementedError
        Conversion from fast_pauli type to Qiskit type isn't supported.
    """
    if isinstance(fast_pauli_type, fp.PauliOp):
        return SparsePauliOp(
            fast_pauli_type.pauli_strings_as_str,
            fast_pauli_type.coeffs,
        )

    elif isinstance(fast_pauli_type, fp.PauliString):
        return Pauli(str(fast_pauli_type))

    else:
        raise NotImplementedError(
            f"Conversion from {type(fast_pauli_type)} to "
            + "Qiskit equivalent isn't supported."
        )



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Test pauli c++ objects against python implementations."""

import itertools as it
from typing import Callable

import pytest

import fast_pauli._fast_pauli as fp
import fast_pauli.pypauli as pp
from tests.conftest import (
    QUBITS_TO_BENCHMARK,
    SAMPLE_STRINGS_LIMIT,
    resolve_parameter_repr,
)

N_OPERATORS_TO_BENCHMARK = 2
N_STATES_TO_BENCHMARK = [16, 64, 256]


@pytest.fixture
def prepared_operators(
    pauli_strings_with_size: Callable,
    pauli_strings_shuffled: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
    qubits: int,
) -> list:
    """Fixture to provide random initialized Pauli operators for testing."""
    operators = []
    for _ in range(N_OPERATORS_TO_BENCHMARK // 2):
        paulis = pauli_strings_shuffled(qubits, limit=SAMPLE_STRINGS_LIMIT)
        coeffs = generate_random_complex(len(paulis))
        operators.append(pauli_op(coeffs, paulis))
    for _ in range(N_OPERATORS_TO_BENCHMARK // 2):
        paulis = pauli_strings_with_size(qubits, limit=SAMPLE_STRINGS_LIMIT)
        coeffs = generate_random_complex(len(paulis))
        operators.append(pauli_op(coeffs, paulis))
    return operators


def benchmark_dense_conversion(operators: list) -> None:
    """Benchmark dense conversion."""
    for op in operators:
        dense_repr = op.to_tensor()  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_op,",
    it.chain(
        [(q, fp.PauliOp) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliOp) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_dense_conversion_n_qubits(
    benchmark: Callable,
    prepared_operators: list,
) -> None:
    """Benchmark dense conversion."""
    benchmark(benchmark_dense_conversion, operators=prepared_operators)


def benchmark_apply(operators: list, states: list) -> None:
    """Benchmark apply method."""
    for op, psi in zip(operators, states):
        result = op.apply(psi)  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_op,",
    it.chain(
        [(q, fp.PauliOp) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliOp) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_apply_n_qubits(
    benchmark: Callable,
    generate_random_complex: Callable,
    prepared_operators: list,
    qubits: int,
) -> None:
    """Benchmark PauliOp multiplication with provided state vector."""
    n_dims = 1 << qubits
    prepared_states = [
        generate_random_complex(n_dims) for _ in range(len(prepared_operators))
    ]

    benchmark(benchmark_apply, operators=prepared_operators, states=prepared_states)


@pytest.mark.parametrize(
    "qubits,states,pauli_op,",
    it.chain(
        [
            (q, n, fp.PauliOp)
            for q in QUBITS_TO_BENCHMARK
            for n in N_STATES_TO_BENCHMARK
        ],
        [
            (q, n, pp.PauliOp)
            for q in QUBITS_TO_BENCHMARK
            for n in N_STATES_TO_BENCHMARK
        ],
    ),
    ids=resolve_parameter_repr,
)
def test_apply_batch_n_qubits_n_states(
    benchmark: Callable,
    prepared_operators: list,
    generate_random_complex: Callable,
    qubits: int,
    states: int,
) -> None:
    """Benchmark PauliOp multiplication with provided set of state vectors."""
    n_dims = 1 << qubits
    prepared_states = [
        generate_random_complex(n_dims, states) for _ in range(len(prepared_operators))
    ]

    benchmark(benchmark_apply, operators=prepared_operators, states=prepared_states)


def benchmark_expectation_value(operators: list, states: list) -> None:
    """Benchmark expectation_value method."""
    for op, psi in zip(operators, states):
        result = op.expectation_value(psi)  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_op,",
    it.chain(
        [(q, fp.PauliOp) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliOp) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_expectation_value_n_qubits(
    benchmark: Callable,
    prepared_operators: list,
    generate_random_complex: Callable,
    qubits: int,
) -> None:
    """Benchmark PauliOp expectation_value with provided state vector."""
    n_dims = 1 << qubits
    prepared_states = [
        generate_random_complex(n_dims) for _ in range(len(prepared_operators))
    ]

    benchmark(
        benchmark_expectation_value,
        operators=prepared_operators,
        states=prepared_states,
    )


@pytest.mark.parametrize(
    "qubits,states,pauli_op,",
    it.chain(
        [
            (q, n, fp.PauliOp)
            for q in QUBITS_TO_BENCHMARK
            for n in N_STATES_TO_BENCHMARK
        ],
        [
            (q, n, pp.PauliOp)
            for q in QUBITS_TO_BENCHMARK
            for n in N_STATES_TO_BENCHMARK
        ],
    ),
    ids=resolve_parameter_repr,
)
def test_expectation_value_batch_n_qubits_n_states(
    benchmark: Callable,
    prepared_operators: list,
    generate_random_complex: Callable,
    qubits: int,
    states: int,
) -> None:
    """Benchmark PauliOp expectation_value with provided set of state vectors."""
    n_dims = 1 << qubits
    prepared_states = [
        generate_random_complex(n_dims, states) for _ in range(len(prepared_operators))
    ]

    benchmark(
        benchmark_expectation_value,
        operators=prepared_operators,
        states=prepared_states,
    )


def benchmark_arithmetic(left_ops: list, right_ops: list) -> None:
    """Benchmark PauliOp arithmetic."""
    for l_op, r_op in zip(left_ops, right_ops):
        p_op1 = l_op + r_op  # noqa: F841
        p_op2 = r_op - l_op  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_op,",
    it.chain(
        [(q, fp.PauliOp) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliOp) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_arithmetic_n_qubits(
    benchmark: Callable,
    prepared_operators: list,
    qubits: int,
) -> None:
    """Benchmark addition and subtraction of PauliOps."""
    benchmark(
        benchmark_arithmetic,
        left_ops=prepared_operators,
        right_ops=reversed(prepared_operators),
    )


def benchmark_inplace_arithmetic(left_ops: list, right_ops: list) -> None:
    """Benchmark PauliOp inplace arithmetic."""
    for l_op, r_op in zip(left_ops, right_ops):
        l_op -= r_op
        r_op += l_op


@pytest.mark.parametrize(
    "qubits,pauli_op,",
    it.chain(
        [(q, fp.PauliOp) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliOp) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_inplace_arithmetic_n_qubits(
    benchmark: Callable,
    prepared_operators: list,
    qubits: int,
) -> None:
    """Benchmark inplace addition and subtraction of PauliOps."""
    benchmark(
        benchmark_inplace_arithmetic,
        left_ops=prepared_operators,
        right_ops=reversed(prepared_operators),
    )


def benchmark_matmul(left_ops: list, right_ops: list) -> None:
    """Benchmark PauliOp/PauliString matrix multiplication."""
    for left, right in zip(left_ops, right_ops):
        op = left @ right  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_op,",
    it.chain(
        [(q, fp.PauliOp) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliOp) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_multiplication_n_qubits(
    benchmark: Callable,
    prepared_operators: list,
    qubits: int,
) -> None:
    """Benchmark matrix multiplication of PauliOps."""
    benchmark(
        benchmark_matmul,
        left_ops=prepared_operators,
        right_ops=reversed(prepared_operators),
    )


@pytest.mark.parametrize(
    "qubits,pauli_op,pauli_string,",
    it.chain(
        [(q, fp.PauliOp, fp.PauliString) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliOp, pp.PauliString) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_multiplication_with_string_n_qubits(
    benchmark: Callable,
    pauli_strings_shuffled: Callable,
    prepared_operators: list,
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
    qubits: int,
) -> None:
    """Benchmark matrix multiplication of PauliOp with PauliString."""
    prepared_strings = list(
        map(
            lambda s: pauli_string(s),
            pauli_strings_shuffled(qubits, limit=len(prepared_operators)),
        )
    )

    benchmark(benchmark_matmul, left_ops=prepared_operators, right_ops=prepared_strings)


if __name__ == "__main__":
    pytest.main()



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################

"""Advanced Qiskit benchmarks for the MVP Release."""

import itertools as it
import os
from typing import Callable

import numpy as np
import pytest
from qiskit.quantum_info import Pauli, SparsePauliOp, Statevector

import fast_pauli as fp

extra_benchmark_mark = pytest.mark.skipif(
    os.getenv("EXTRA_BENCHMARKS", "false").lower() != "true",
    reason="ENV variable EXTRA_BENCHMARKS is not set to true",
)


################################################################################
# Pauli String Apply
################################################################################
qiskit_params_sm = [
    pytest.param(nq, m) for nq, m in it.product([2, 4, 6, 8, 10], ["qiskit"])
]
qiskit_params_big = [
    pytest.param(nq, m, marks=extra_benchmark_mark)
    for nq, m in it.product([12, 14], ["qiskit"])
]

fast_pauli_params_sm = [
    pytest.param(nq, m) for nq, m in it.product([2, 4, 6, 8, 10], ["fast_pauli"])
]

fast_pauli_params_big = [
    pytest.param(nq, m, marks=extra_benchmark_mark)
    for nq, m in it.product([12, 14, 16, 18, 20, 22, 24], ["fast_pauli"])
]


@pytest.mark.parametrize(
    "n_qubits, method",
    qiskit_params_sm + qiskit_params_big + fast_pauli_params_sm + fast_pauli_params_big,
)
def test_pauli_string_apply(benchmark: Callable, n_qubits: int, method: str) -> None:
    """Benchmark applying a Pauli String to a state."""
    np.random.seed(18)

    all_pauli_strings = ["".join(s) for s in it.product("IXYZ", repeat=n_qubits)]
    pauli_string = np.random.choice(all_pauli_strings, size=1)[0]
    print(f"n_qubits: {n_qubits}, method: {method}, pauli_string: {pauli_string}")

    if method == "qiskit":
        v = Statevector(np.random.random(2**n_qubits).astype(np.complex128))
        ps = Pauli(pauli_string)

        def f() -> None:
            v.evolve(ps)

    else:
        v = np.random.random(2**n_qubits).astype(np.complex128)
        ps = fp.PauliString(pauli_string)

        def f() -> None:
            ps.apply(v)

    benchmark(f)


################################################################################
# Pauli Op Apply
################################################################################
qiskit_params_sm = [
    pytest.param(nq, ns, m)
    for nq, ns, m in it.product([2, 8, 10], [10, 100, 1000], ["qiskit"])
]
qiskit_params_big = [
    pytest.param(nq, ns, m, marks=extra_benchmark_mark)
    for nq, ns, m in it.product([12, 13, 14, 15], [10, 100, 1000], ["qiskit"])
]

fast_pauli_params_sm = [
    pytest.param(nq, ns, m)
    for nq, ns, m in it.product([2, 8, 10], [10, 100, 1000], ["fast_pauli"])
]

fast_pauli_params_big = [
    pytest.param(nq, ns, m, marks=extra_benchmark_mark)
    for nq, ns, m in it.product(
        [12, 13, 14, 15, 16, 18], [10, 100, 1000], ["fast_pauli"]
    )
]


@pytest.mark.parametrize(
    "n_qubits, n_pauli_strings, method",
    qiskit_params_sm + qiskit_params_big + fast_pauli_params_sm + fast_pauli_params_big,
)
def test_pauli_op_apply(
    benchmark: Callable,
    pauli_strings_with_size: Callable,
    n_qubits: int,
    n_pauli_strings: int,
    method: str,
) -> None:
    """Benchmark applying a Pauli Operator (in the general sense) to a state."""
    np.random.seed(18)

    all_pauli_strings = ["".join(s) for s in it.product("IXYZ", repeat=n_qubits)]
    pauli_strings = np.random.choice(
        all_pauli_strings, size=n_pauli_strings, replace=True
    )
    print(
        f"n_qubits: {n_qubits}, n_pauli_strings:"
        + f" {n_pauli_strings} ?= {len(pauli_strings)}, method: {method}"
    )

    v_np = np.random.rand(2**n_qubits).astype(np.complex128)

    if method == "qiskit":
        v = Statevector(v_np)
        op = SparsePauliOp(pauli_strings, coeffs=np.ones(len(pauli_strings)))

        def f() -> None:
            v.evolve(op)
    else:
        op = fp.PauliOp(np.ones(len(pauli_strings)), pauli_strings)

        def f() -> None:
            op.apply(v_np)

    benchmark(f)


################################################################################
# Pauli Op Expectation Value Batch
################################################################################
qiskit_params_sm = [
    pytest.param(nq, npauli, nstate, m)
    for nq, npauli, nstate, m in it.product([2, 8, 10, 12], [1024], [1, 10], ["qiskit"])
]

qiskit_params_big = [
    pytest.param(nq, npauli, nstate, m, marks=extra_benchmark_mark)
    for nq, npauli, nstate, m in it.product(
        [2, 8, 10, 12, 14, 16], [1024], [1, 10, 100], ["qiskit"]
    )
]

fast_pauli_params_sm = [
    pytest.param(nq, npauli, nstate, m)
    for nq, npauli, nstate, m in it.product(
        [2, 8, 10, 12], [1024], [1, 10, 100, 1000], ["fast_pauli"]
    )
]

fast_pauli_params_big = [
    pytest.param(nq, npauli, nstate, m, marks=extra_benchmark_mark)
    for nq, npauli, nstate, m in it.product(
        [14, 15, 16], [1024], [1, 10, 100, 1000], ["fast_pauli"]
    )
]


@pytest.mark.parametrize(
    "n_qubits, n_pauli_strings, n_states, method",
    qiskit_params_sm + qiskit_params_big + fast_pauli_params_sm + fast_pauli_params_big,
)
def test_pauli_op_expectation_value_batch(
    benchmark: Callable, n_qubits: int, n_pauli_strings: int, n_states: int, method: str
) -> None:
    """Benchmark the expectation value of a Pauli Operator to a batch of states."""
    np.random.seed(18)
    all_pauli_strings = ["".join(s) for s in it.product("IXYZ", repeat=n_qubits)]
    pauli_strings = np.random.choice(
        all_pauli_strings, size=n_pauli_strings, replace=True
    )
    print(
        f"n_qubits: {n_qubits}, n_pauli_strings:"
        + f" {n_pauli_strings}, n_states: {n_states}, "
        + f"method: {method}, "
        + f"len(pauli_strings): {len(pauli_strings)}"
    )

    v_np = np.random.rand(2**n_qubits, n_states).astype(np.complex128)

    if method == "qiskit":
        op = SparsePauliOp(pauli_strings, coeffs=np.ones(len(pauli_strings)))
        v_np = v_np.T.copy()
        state_batch = [Statevector(v) for v in v_np]

        def f() -> None:
            results = [s.expectation_value(op) for s in state_batch]  # noqa: F841

    else:
        op = fp.PauliOp(np.ones(len(pauli_strings)), pauli_strings)

        def f() -> None:
            op.expectation_value(v_np)

    benchmark(f)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "fast_pauli"
copyright = "2024, Qognitive, Inc."
author = "James E. T. Smith, Eugene Rublenko, Alex Lerner, Sebastien Roy, Jeffrey Berger"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["breathe", "sphinx.ext.autodoc", "sphinx_copybutton", "sphinx.ext.mathjax", "sphinx.ext.napoleon"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_favicon = "logo/Sigma-logo-small.png"
html_static_path = []  # type: ignore


# Breathe configuration
breathe_projects = {
    "fast_pauli": "xml",
}
breathe_default_project = "fast_pauli"



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Separate file with helper functions for Pauli matrices."""

import numpy as np


def pauli_matrices() -> dict[str | int, np.ndarray]:
    """Provide a dictionary containing the Pauli matrices.

    The Pauli matrices are a set of 2x2 matrices commonly used in quantum mechanics.
    This function returns a dictionary with keys representing the matrix names
    ("I", "X", "Y", "Z", 0, 1, 2, 3) and values representing the corresponding matrix.

    Returns
    -------
        dict: A dictionary containing the Pauli matrices.

    """
    s0 = np.array([[1, 0], [0, 1]], dtype=np.complex128)
    s1 = np.array([[0, 1], [1, 0]], dtype=np.complex128)
    s2 = np.array([[0, -1j], [1j, 0]], dtype=np.complex128)
    s3 = np.array([[1, 0], [0, -1]], dtype=np.complex128)
    return {"I": s0, "X": s1, "Y": s2, "Z": s3, 0: s0, 1: s1, 2: s2, 3: s3}


def naive_pauli_converter(string: str) -> np.ndarray:
    """Convert Pauli string representation into corresponding dense matrix.

    Parameters
    ----------
    string : str
        The string representation of Pauli string.

    Returns
    -------
    np.ndarray
        The matrix representation of the Pauli string.

    """
    paulis = pauli_matrices()
    matrix = paulis[string[-1]]
    for p in reversed(string[:-1]):
        matrix = np.kron(paulis[p], matrix)
    return matrix


def naive_pauli_operator(coeffs: list[np.complex128], strings: list[str]) -> np.ndarray:
    """Convert list of Pauli strings and coefficients into corresponding dense matrix.

    Parameters
    ----------
    coeffs : list[float]
        The list of coefficients corresponding to the Pauli strings.
    strings : list[str]
        The list of string representations of Pauli strings.

    Returns
    -------
    np.ndarray
        The matrix representation of the Pauli operator.

    """
    if len(coeffs) != len(strings):
        raise ValueError("Length of coefficients and strings must match.")
    return np.sum(
        [
            coeff * naive_pauli_converter(string)
            for string, coeff in zip(strings, coeffs)
        ],
        axis=0,
    )



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""pytest configuration file for fast_pauli tests."""

import itertools as it
from typing import Any, Callable

import numpy as np
import pytest

from fast_pauli.pypauli.helpers import pauli_matrices

### TEST CONSTANTS ###

SAMPLE_STRINGS_LIMIT = 256
QUBITS_TO_BENCHMARK = [2, 8]

LARGE_QUBITS = [14, 17, 20]


### TEST FIXTURES ###


@pytest.fixture
def paulis() -> dict[str | int, np.ndarray]:
    """Fixture to provide dict with Pauli matrices."""
    return pauli_matrices()


@pytest.fixture
def sample_pauli_strings() -> list[str]:
    """Fixture to provide sample Pauli strings for testing."""
    strings = it.chain(
        ["I", "X", "Y", "Z"],
        it.product("IXYZ", repeat=2),
        it.product("IXYZ", repeat=3),
        ["XYZXYZ", "ZZZIII", "XYIZXYZ", "XXIYYIZZ", "ZIXIZYXX"],
    )
    return list(map("".join, strings))


@pytest.fixture
def pauli_strings_with_size() -> Callable:
    """Fixture to provide ordered Pauli strings of desired size for testing."""

    def generate_paulis(size: int, limit: int = 1_000) -> list[str]:
        strings: list[str] = []
        for s in it.product("IXYZ", repeat=size):
            if limit and len(strings) >= limit:
                break
            strings.append("".join(s))
        return strings

    return generate_paulis


@pytest.fixture
def pauli_strings_shuffled() -> Callable:
    """Fixture to provide shuffled Pauli strings of desired size for testing."""

    def generate_paulis(size: int, limit: int = 1_000) -> list[str]:
        # pretty sloppy implementation that uses global RNG for now
        all_strings = ["".join(s) for s in it.product("IXYZ", repeat=size)]
        n_choices = limit if limit and limit <= len(all_strings) else len(all_strings)
        strings: list[str] = np.random.choice(
            all_strings, size=n_choices, replace=False
        ).tolist()
        return strings

    return generate_paulis


@pytest.fixture(scope="function")
def generate_random_complex(rng_seed: int = 321) -> np.ndarray:
    """Fixture to generate random complex numpy array with desired shape."""
    rng = np.random.default_rng(rng_seed)
    return lambda *shape: rng.random(shape) + 1j * rng.random(shape)


### TEST UTILITIES ###


def resolve_parameter_repr(val: Any) -> Any | str:
    """Regular function to resolve representation for pytest parametrization.

    Currently, the main purpose is to automatically distinct cpp and py implementations
    for test logging
    """
    module_name: str | None = getattr(val, "__module__", None)
    if module_name is None:
        return val
    elif "_fast_pauli" in module_name:
        return val.__qualname__ + "-cpp"
    elif "pypauli" in module_name:
        return val.__qualname__ + "-py"
    return val



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Efficient operations on Pauli string using numpy."""

from __future__ import annotations

from typing import Any

import numpy as np


class PauliString:
    """Class representing a Pauli string with efficient operations."""

    def __init__(self, string: str) -> None:
        """Initialize the PauliString object.

        Args:
        ----
            string: The input string representing the Pauli string.
            Must contain only I, X, Y, Z characters.

        """
        if not all([c in "IXYZ" for c in string]):
            raise ValueError(f"Invalid pauli string '{string}'")

        self.string = string
        self.n_qubits = len(string)
        self.dim = 1 << len(string) if string else 0
        self.weight = len(string) - string.count("I")

    def __str__(self) -> str:
        """Convert pauli string object to regular string.

        Returns
        -------
            str: Plain pauli string.
        """
        return self.string

    def __repr__(self) -> str:
        """Get a string representation of the PauliString object.

        Returns
        -------
            str: A string representation of the PauliString object.
        """
        return f'PauliString("{self.string}")'

    def copy(self) -> PauliString:
        """Return a copy of the Pauli string."""
        return PauliString(self.string)

    def to_tensor(self) -> np.ndarray:
        """Return the dense matrix representation of the Pauli string."""
        columns, values = compose_sparse_pauli(self.string)

        matrix = np.zeros((columns.size, values.size), dtype=np.complex128)
        matrix[np.arange(columns.size), columns] = values
        return matrix

    def apply(self, states: np.ndarray, coeff: np.complex128 = 1.0) -> np.ndarray:
        """Efficient multiplication of Pauli string with a given state.

        Args:
        ----
            state: The input state as a numpy array.
            coeff: Multiplication factor to apply to the PauliString

        Returns
        -------
            The result of multiplying the Pauli string with the state.

        """
        if states.shape[0] != self.dim or states.ndim > 2:
            raise ValueError(f"Provided state has inconsistent shape {states.shape}")

        columns, values = compose_sparse_pauli(self.string)
        values *= coeff

        if states.ndim == 2:
            return values[:, np.newaxis] * states[columns]
        else:
            return values * states[columns]

    def expectation_value(self, state: np.ndarray) -> np.complex128 | np.ndarray:
        """Compute the expected value of Pauli string for a given state.

        Args:
        ----
            state: The input state as a numpy array.
            coeff: Multiplication factor to apply to the PauliString

        Returns
        -------
            The expected value of the Pauli string with the state.

        """
        return np.multiply(state.conj(), self.apply(state)).sum(axis=0)

    def multiply(self, rhs: PauliString) -> tuple[np.complex128, PauliString]:
        """Matrix multiplication of two Pauli strings.

        Args:
        ----
            rhs: The other PauliString object to multiply with.

        Returns
        -------
            tuple containing the multiplication factor and resulting PauliString object.
        """
        if self.dim != rhs.dim:
            raise ValueError("Pauli strings must have the same length")

        phase = 1 + 0j
        p_str = []
        for p_left, p_right in zip(self.string, rhs.string):
            if p_left == p_right:
                p_str.append("I")
            elif p_left == "I":
                p_str.append(p_right)
            elif p_right == "I":
                p_str.append(p_left)
            elif p_left == "X" and p_right == "Y":
                p_str.append("Z")
                phase *= 1j
            elif p_left == "X" and p_right == "Z":
                p_str.append("Y")
                phase *= -1j
            elif p_left == "Y" and p_right == "X":
                p_str.append("Z")
                phase *= -1j
            elif p_left == "Y" and p_right == "Z":
                p_str.append("X")
                phase *= 1j
            elif p_left == "Z" and p_right == "Y":
                p_str.append("X")
                phase *= -1j
            elif p_left == "Z" and p_right == "X":
                p_str.append("Y")
                phase *= 1j
        return phase, PauliString("".join(p_str))

    def __matmul__(self, rhs: PauliString) -> tuple[np.complex128, PauliString]:
        """Matrix multiplication of two Pauli strings.

        Args:
        ----
            rhs: The other PauliString object to multiply with.

        Returns
        -------
            tuple containing the multiplication factor and resulting PauliString object.
        """
        if isinstance(rhs, PauliString):
            return self.multiply(rhs)
        else:
            # point python to __rmatmul__ method of PauliOp in case if rhs is PauliOp
            return NotImplemented

    def __add__(self, rhs: PauliString):  # type: ignore
        """Add two PauliString objects.

        Args:
        ----
            rhs: The other PauliString object to add.

        Returns
        -------
            PauliOp: Pauli Op holding the sum of the two PauliStrings.
        """
        if self.dim != rhs.dim:
            raise ValueError("Pauli strings must have the same length")

        from .pauli_op import PauliOp

        # reroute this to PauliOp operator since the resulting object is PauliOp anyway
        return PauliOp([1.0], [self]) + rhs

    def __sub__(self, rhs: PauliString):  # type: ignore
        """Subtract PauliString from the current one.

        Args:
        ----
            rhs: The other PauliString object to subtract.

        Returns
        -------
            PauliOp: Pauli Op holding the difference of the two PauliStrings.
        """
        if self.dim != rhs.dim:
            raise ValueError("Pauli strings must have the same length")

        from .pauli_op import PauliOp

        # reroute this to PauliOp operator since the resulting object is PauliOp anyway
        return PauliOp([1.0], [self]) - rhs

    def __eq__(self, other: PauliString | Any) -> bool:
        """Check if two PauliString objects are equal.

        Args:
        ----
            other: The other PauliString object to compare with.

        Returns
        -------
            bool: True if the PauliString objects are equal, False otherwise.
        """
        if not isinstance(other, PauliString):
            raise NotImplementedError("Cannot compare PauliString with other types")
        return self.string == other.string

    def __ne__(self, other: PauliString | Any) -> bool:
        """Check if two PauliString objects are not equal.

        Args:
        ----
            other: The other PauliString object to compare with.

        Returns
        -------
            bool: True if the PauliString objects are not equal, False otherwise.
        """
        return not (self == other)

    def __hash__(self) -> int:
        """Return the hash value of the PauliString object."""
        return hash(self.string)


def compose_sparse_pauli(string: str) -> tuple[np.ndarray, np.ndarray]:
    """Produce sparse representation of the pauli string.

    Args:
    ----
        string: The input string representing the Pauli string.
        Must contain only I, X, Y, Z characters.

    Returns
    -------
        A tuple containing the column numbers and values of the sparse matrix.

    """
    n_qubits = len(string)
    n_vals = 1 << n_qubits if string else 0
    n_ys = string.count("Y")

    # initialize cols array with zeros as we need first element to be 0
    cols = np.zeros(n_vals, dtype=np.int32)
    vals = np.empty(n_vals, dtype=np.complex128)

    if not n_vals:
        return cols, vals

    for p in string:
        cols[0] <<= 1
        if p == "X" or p == "Y":
            cols[0] += 1

    match n_ys % 4:
        case 0:
            vals[0] = 1.0
        case 1:
            vals[0] = -1.0j
        case 2:
            vals[0] = -1.0
        case 3:
            vals[0] = 1.0j

    for q in range(n_qubits):
        p = string[n_qubits - q - 1]
        pow_of_two = 1 << q

        new_slice = slice(pow_of_two, 2 * pow_of_two)
        old_slice = slice(0, pow_of_two)

        match p:
            case "I":
                cols[new_slice] = cols[old_slice] + pow_of_two
                vals[new_slice] = vals[old_slice]
            case "X":
                cols[new_slice] = cols[old_slice] - pow_of_two
                vals[new_slice] = vals[old_slice]
            case "Y":
                cols[new_slice] = cols[old_slice] - pow_of_two
                vals[new_slice] = -vals[old_slice]
            case "Z":
                cols[new_slice] = cols[old_slice] + pow_of_two
                vals[new_slice] = -vals[old_slice]

    return cols, vals


def compose_sparse_diag_pauli(string: str) -> np.ndarray:
    """Produce sparse representation of diagonal pauli string.

    Args:
    ----
        string: A Pauli string containing only 'I' and 'Z' characters.

    Returns
    -------
        np.ndarray: diagonal values from resulting sparse matrix.

    """
    if "X" in string or "Y" in string:
        raise ValueError("Pauli string must contain only I and Z characters")

    n_qubits = len(string)
    n_vals = 1 << n_qubits

    # initialize vals array with ones as we need first element to be 1
    vals = np.ones(n_vals, dtype=np.complex128)

    for q in range(n_qubits):
        p = string[n_qubits - q - 1]
        pow_of_two = 1 << q

        new_slice = slice(pow_of_two, 2 * pow_of_two)
        old_slice = slice(0, pow_of_two)

        match p:
            case "I":
                vals[new_slice] = vals[old_slice]
            case "Z":
                vals[new_slice] = -vals[old_slice]

    return vals



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


""":code:`fast_pauli` is a Python package with C++ backend, optimized for operations on Pauli Matrices and Pauli Strings."""  # noqa: E501

from ._fast_pauli import (  # noqa: F401
    Pauli,
    PauliOp,
    PauliString,
    SummedPauliOp,
    helpers,
)
from .qiskit_helpers import from_qiskit, to_qiskit  # noqa: F401



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Benchmark fast_pauli operations with qiskit operations."""

import itertools as it
from typing import Callable, List

import pytest
from qiskit.quantum_info import Pauli, SparsePauliOp

import fast_pauli as fp
from tests.conftest import (
    QUBITS_TO_BENCHMARK,
    SAMPLE_STRINGS_LIMIT,
    resolve_parameter_repr,
)

N_OPERATORS_TO_BENCHMARK = 64


@pytest.fixture
def prepared_paulis(
    pauli_strings_with_size: Callable,
    pauli_class: type[fp.PauliString] | type[Pauli],
    qubits: int,
) -> list:
    """Fixture to provide initialized Pauli strings for testing."""
    return list(
        map(
            lambda s: pauli_class(s),
            pauli_strings_with_size(qubits, limit=SAMPLE_STRINGS_LIMIT),
        )
    )


@pytest.fixture
def prepared_pauliops(
    pauli_strings_with_size: Callable,
    pauli_strings_shuffled: Callable,
    generate_random_complex: Callable,
    pauli_class: type[fp.PauliOp] | type[SparsePauliOp],
    qubits: int,
) -> list:
    """Fixture to provide random initialized Pauli operators for testing."""
    operators = []
    for _ in range(N_OPERATORS_TO_BENCHMARK // 2):
        paulis = pauli_strings_shuffled(qubits, limit=SAMPLE_STRINGS_LIMIT)
        coeffs = generate_random_complex(len(paulis))
        if pauli_class == fp.PauliOp:
            operators.append(fp.PauliOp(coeffs, paulis))
        elif pauli_class == SparsePauliOp:
            operators.append(SparsePauliOp(paulis, coeffs=coeffs))
        else:
            raise ValueError(f"Unknown Pauli class: {pauli_class}")
    for _ in range(N_OPERATORS_TO_BENCHMARK // 2):
        paulis = pauli_strings_with_size(qubits, limit=SAMPLE_STRINGS_LIMIT)
        coeffs = generate_random_complex(len(paulis))
        if pauli_class == fp.PauliOp:
            operators.append(fp.PauliOp(coeffs, paulis))
        else:
            operators.append(SparsePauliOp(paulis, coeffs=coeffs))
    return operators


def benchmark_sum(paulis: List) -> None:
    """Benchmark addition."""
    for p0, p1 in zip(paulis, reversed(paulis)):
        _ = p0 + p1  # noqa: F841


def benchmark_dot(paulis: List) -> None:
    """Benchmark multiplication."""
    for p0, p1 in zip(paulis, reversed(paulis)):
        _ = p0 @ p1  # noqa: F841


def benchmark_qiskit_to_dense(paulis: List) -> None:
    """Benchmark Qiskit SparsePauliOp.to_matrix()."""
    for p in paulis:
        _ = p.to_matrix()  # noqa: F841


def benchmark_fp_to_dense(paulis: List) -> None:
    """Benchmark FastPauli PauliOp.to_tensor()."""
    for p in paulis:
        _ = p.to_tensor()  # noqa: F841


def benchmark_fp_squared(paulis: List) -> None:
    """Benchmark squaring for Fast_Pauli."""
    for p in paulis:
        _ = p @ p  # noqa: F841


def benchmark_qiskit_squared(paulis: List) -> None:
    """Benchmark squaring for Qiskit."""
    for p in paulis:
        _ = p.power(2)


@pytest.mark.parametrize(
    "pauli_class, qubits",
    it.product(
        [fp.PauliOp, SparsePauliOp],
        QUBITS_TO_BENCHMARK,
    ),
    ids=resolve_parameter_repr,
)
def test_sum_pauliops(
    benchmark: Callable,
    prepared_pauliops: List,
) -> None:
    """Benchmark addition for Pauli operators."""
    benchmark(benchmark_sum, paulis=prepared_pauliops)


@pytest.mark.parametrize(
    "pauli_class, qubits",
    it.product(
        [fp.PauliString, Pauli],
        QUBITS_TO_BENCHMARK,
    ),
    ids=resolve_parameter_repr,
)
def test_dot_paulis(
    benchmark: Callable,
    prepared_paulis: List,
) -> None:
    """Benchmark dot product for fp.PauliString and Qiskit Pauli."""
    benchmark(benchmark_dot, paulis=prepared_paulis)


@pytest.mark.parametrize(
    "pauli_class, qubits",
    it.product(
        [fp.PauliOp, SparsePauliOp],
        QUBITS_TO_BENCHMARK,
    ),
    ids=resolve_parameter_repr,
)
def test_dot_pauliops(
    benchmark: Callable,
    prepared_pauliops: List,
) -> None:
    """Benchmark dot product for fp.PauliOp and Qiskit SparsePauliOp."""
    benchmark(benchmark_dot, paulis=prepared_pauliops)


@pytest.mark.parametrize(
    "pauli_class, qubits",
    it.product(
        [fp.PauliString, Pauli],
        QUBITS_TO_BENCHMARK,
    ),
    ids=resolve_parameter_repr,
)
def test_to_dense_paulis(
    benchmark: Callable,
    prepared_paulis: List,
) -> None:
    """Benchmark conversion to dense matrix for fp.PauliString and Qiskit Pauli."""
    # Qiskit and Fast_Pauli use differently named methods to convert to dense matrix
    if type(prepared_paulis[0]) is fp.PauliString:
        benchmark(benchmark_fp_to_dense, paulis=prepared_paulis)
    elif type(prepared_paulis[0]) is Pauli:
        benchmark(benchmark_qiskit_to_dense, paulis=prepared_paulis)
    else:
        raise ValueError(f"Unknown Pauli class: {type(prepared_paulis[0])}")


@pytest.mark.parametrize(
    "pauli_class, qubits",
    it.product(
        [fp.PauliOp, SparsePauliOp],
        QUBITS_TO_BENCHMARK,
    ),
    ids=resolve_parameter_repr,
)
def test_to_dense_pauliops(
    benchmark: Callable,
    prepared_pauliops: List,
) -> None:
    """Benchmark conversion to dense matrix for fp.PauliOp and Qiskit SparsePauliOp."""
    # Qiskit and Fast_Pauli use differently named methods to convert to dense matrix
    if type(prepared_pauliops[0]) is fp.PauliOp:
        benchmark(benchmark_fp_to_dense, paulis=prepared_pauliops)
    elif type(prepared_pauliops[0]) is SparsePauliOp:
        benchmark(benchmark_qiskit_to_dense, paulis=prepared_pauliops)
    else:
        raise ValueError(f"Unknown Pauli class: {type(prepared_pauliops[0])}")


@pytest.mark.parametrize(
    "pauli_class, qubits",
    it.product(
        [fp.PauliString, Pauli],
        QUBITS_TO_BENCHMARK,
    ),
    ids=resolve_parameter_repr,
)
def test_square_paulis(
    benchmark: Callable,
    prepared_paulis: List,
) -> None:
    """Benchmark squaring for fp.PauliString and Qiskit Pauli."""
    if type(prepared_paulis[0]) is fp.PauliString:
        benchmark(benchmark_fp_squared, paulis=prepared_paulis)
    elif type(prepared_paulis[0]) is Pauli:
        benchmark(benchmark_qiskit_squared, paulis=prepared_paulis)
    else:
        raise ValueError(f"Unknown Pauli class: {type(prepared_paulis[0])}")


@pytest.mark.parametrize(
    "pauli_class, qubits",
    it.product(
        [fp.PauliOp, SparsePauliOp],
        QUBITS_TO_BENCHMARK,
    ),
    ids=resolve_parameter_repr,
)
def test_square_pauliops(
    benchmark: Callable,
    prepared_pauliops: List,
) -> None:
    """Benchmark squaring for fp.PauliOp and Qiskit SparsePauliOp."""
    if type(prepared_pauliops[0]) is fp.PauliOp:
        benchmark(benchmark_fp_squared, paulis=prepared_pauliops)
    elif type(prepared_pauliops[0]) is SparsePauliOp:
        benchmark(benchmark_qiskit_squared, paulis=prepared_pauliops)
    else:
        raise ValueError(f"Unknown Pauli class: {type(prepared_pauliops[0])}")


def benchmark_pauliop_pauli_mult(pauli_ops: List, pauli_strings: List) -> None:
    """Benchmark fast_pauli.PauliOp @ fast_pauli.PauliString."""
    for pOp, ps in zip(pauli_ops, pauli_strings):
        _ = pOp @ ps  # noqa: F841


def benchmark_qiskit_pauliop_pauli_mult(pauli_ops: List, pauli_strings: List) -> None:
    """Benchmark Qiskit SparsePauliOp @ Qiskit Pauli."""
    for pOp, ps in zip(pauli_ops, pauli_strings):
        _ = pOp.dot(ps)  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_class,pauli_string_class,",
    it.chain(
        [(q, fp.PauliOp, fp.PauliString) for q in QUBITS_TO_BENCHMARK],
        [(q, SparsePauliOp, Pauli) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_pauliop_pauli_mult(
    benchmark: Callable,
    pauli_class: type[fp.PauliOp] | type[SparsePauliOp],
    pauli_string_class: type[fp.PauliString] | type[Pauli],
    prepared_pauliops: List,
    qubits: int,
    pauli_strings_with_size: Callable,
) -> None:
    """Benchmark PauliOp @ PauliString vs Qiskit SparsePauliOp @ Pauli."""
    prepared_strings = [
        pauli_string_class(s)
        for s in pauli_strings_with_size(qubits, limit=SAMPLE_STRINGS_LIMIT)
    ]
    if pauli_class == fp.PauliOp:
        benchmark(
            benchmark_pauliop_pauli_mult,
            pauli_ops=prepared_pauliops,
            pauli_strings=prepared_strings,
        )
    else:
        benchmark(
            benchmark_qiskit_pauliop_pauli_mult,
            pauli_ops=prepared_pauliops,
            pauli_strings=prepared_strings,
        )



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""PauliOp implementation using numpy."""

from __future__ import annotations

import numpy as np

from fast_pauli.pypauli.pauli_string import PauliString, compose_sparse_pauli


class PauliOp:
    """Class representing a Pauli operator with efficient operations."""

    def __init__(
        self, coefficients: list | np.ndarray, strings: list[str] | list[PauliString]
    ) -> None:
        """Initialize the PauliOp object.

        Args:
        ----
            coefficients: The input coefficients for the Pauli Operator
            strings: Set of Pauli Strings to compose an operator.

        """
        if len(coefficients) != len(strings):
            raise ValueError("coefficients and strings must have the same size.")

        self.coeffs = np.array(coefficients, dtype=np.complex128, copy=True)
        self.pauli_strings = [
            PauliString(ps) if isinstance(ps, str) else ps for ps in strings
        ]

        if self.pauli_strings:
            self.n_qubits = self.pauli_strings[0].n_qubits
            self.dim = self.pauli_strings[0].dim
            for ps in self.pauli_strings[1:]:
                if ps.n_qubits != self.n_qubits:
                    raise ValueError("All Pauli strings must have the same length.")
        else:
            self.n_qubits = 0
            self.dim = 0

    @property
    def pauli_strings_as_str(self) -> list[str]:
        """Get the list of Pauli strings."""
        return [str(ps) for ps in self.pauli_strings]

    @property
    def n_pauli_strings(self) -> int:
        """Get the number of Pauli strings."""
        return len(self.pauli_strings)

    def copy(self) -> PauliOp:
        """Return a copy of the Pauli operator."""
        # TODO proper interface for deep&shallow copying of PauliOp and PauliString
        return PauliOp(self.coeffs, self.pauli_strings_as_str)

    def scale(self, factors: complex | np.ndarray) -> None:
        """Scale each individual term by a factor.

        Args:
        ----
            factors: The factors to scale Pauli operator's coefficients by.
        """
        if isinstance(factors, (complex, float, int)):
            self.coeffs *= factors
        else:
            if factors.shape != self.coeffs.shape:
                raise ValueError("Invalid shape for the scaling factors.")
            self.coeffs *= factors

    def to_tensor(self) -> np.ndarray:
        """Return the dense matrix representation of the Pauli operator."""
        matrix = np.zeros((self.dim, self.dim), dtype=np.complex128)

        for i, ps in enumerate(self.pauli_strings):
            columns, values = compose_sparse_pauli(ps.string)
            matrix[np.arange(self.dim), columns] += self.coeffs[i] * values

        return matrix

    def apply(self, states: np.ndarray) -> np.ndarray:
        """Efficient multiplication of Pauli operator with a given state.

        Args:
        ----
            state: The input state as a numpy array.

        Returns
        -------
            The result of multiplying the Pauli operator with the state.

        """
        if states.shape[0] != self.dim:
            raise ValueError(f"Provided state has inconsistent shape {states.shape}")

        result = np.zeros_like(states, dtype=np.complex128)
        for c, ps in zip(self.coeffs, self.pauli_strings):
            result += ps.apply(states, c)
        return result

    def expectation_value(self, states: np.ndarray) -> np.complex128 | np.ndarray:
        """Compute the expected value of the operator for the given state.

        Args:
        ----
            state: The input state as a numpy array.

        Returns
        -------
            The expected value of the operator with the given state.

        """
        result = np.zeros(
            1 if len(states.shape) == 1 else states.shape[-1], dtype=np.complex128
        )
        for c, ps in zip(self.coeffs, self.pauli_strings):
            result += c * ps.expectation_value(states)
        return result

    def __matmul__(self, rhs: PauliOp | PauliString) -> PauliOp:
        """Matrix multiplication with other PauliOp or PauliString object.

        Args:
        ----
            rhs: The Pauli operator or string

        Returns
        -------
            The result of multiplication
        """
        if self.dim != rhs.dim:
            raise ValueError("Pauli operators must have the same dimension.")

        if isinstance(rhs, PauliString):
            new_strings = []
            new_coeffs = self.coeffs.copy()

            for i, p_str in enumerate(self.pauli_strings):
                phase, new_p_str = p_str @ rhs
                new_strings.append(new_p_str)
                new_coeffs[i] *= phase

        elif isinstance(rhs, PauliOp):
            new_strings = []
            new_coeffs = []
            dedupe_strings: dict[PauliString, int] = {}

            for i, l_str in enumerate(self.pauli_strings):
                for j, r_str in enumerate(rhs.pauli_strings):
                    phase, new_p_str = l_str @ r_str
                    coeff_ij = phase * self.coeffs[i] * rhs.coeffs[j]

                    if new_p_str in dedupe_strings:
                        new_coeffs[dedupe_strings[new_p_str]] += coeff_ij
                    else:
                        dedupe_strings[new_p_str] = len(new_strings)
                        new_strings.append(new_p_str)
                        new_coeffs.append(coeff_ij)

            new_coeffs = np.array(new_coeffs, dtype=np.complex128)
        else:
            raise ValueError("Invalid type for the operand.")

        return PauliOp(new_coeffs, new_strings)

    def __rmatmul__(self, lhs: PauliString) -> PauliOp:
        """Matrix multiplication with PauliString on the left.

        Args:
        ----
            lhs: The Pauli string to multiply with.

        Returns
        -------
            The result of multiplication
        """
        if self.dim != lhs.dim:
            raise ValueError("Pauli operators must have the same dimension.")

        if isinstance(lhs, PauliString):
            new_strings = []
            new_coeffs = self.coeffs.copy()

            for i, p_str in enumerate(self.pauli_strings):
                phase, new_p_str = lhs @ p_str
                new_strings.append(new_p_str)
                new_coeffs[i] *= phase
        else:
            raise ValueError("Invalid type for the operand.")

        return PauliOp(new_coeffs, new_strings)

    def __mul__(self, rhs: complex) -> PauliOp:
        """Multiply the PauliOp with a scalar.

        Args:
        ----
            rhs: The scalar to multiply with.

        Returns
        -------
            PauliOp: Pauli Op holding the product of the PauliOp and the scalar.
        """
        return PauliOp(self.coeffs * rhs, self.pauli_strings_as_str)

    def __rmul__(self, lhs: complex) -> PauliOp:
        """Multiply the PauliOp with a scalar.

        Args:
        ----
            lhs: The scalar to multiply with.

        Returns
        -------
            PauliOp: Pauli Op holding the product of the PauliOp and the scalar.
        """
        return self * lhs

    def __imul__(self, other: complex) -> PauliOp:
        """Multiply the PauliOp with a scalar.

        Args:
        ----
            other: The scalar to multiply with.
        """
        self.scale(other)
        return self

    def __add__(self, rhs: PauliOp | PauliString) -> PauliOp:
        """Add two PauliOp objects together.

        Args:
        ----
            rhs: The other PauliOp object to add.

        Returns
        -------
            PauliOp: Pauli Op holding the sum of the two PauliOps.
        """
        res_op = self.copy()
        res_op.extend(rhs.copy())
        return res_op

    def __iadd__(self, rhs: PauliOp | PauliString) -> PauliOp:
        """Add another PauliOp to the current one.

        Args:
        ----
            rhs: The Pauli operator to be added.
        """
        self.extend(rhs.copy())
        return self

    def __sub__(self, rhs: PauliOp | PauliString) -> PauliOp:
        """Subtract another PauliOp from the current one.

        Args:
        ----
            rhs: The Pauli operator to be subtracted.

        Returns
        -------
            PauliOp: Pauli Op holding the difference of the two PauliOps.
        """
        res_op = self.copy()
        res_op.extend(rhs.copy(), -1.0)
        return res_op

    def __isub__(self, rhs: PauliOp | PauliString) -> PauliOp:
        """Subtract another PauliOp from the current one.

        Args:
        ----
            rhs: The Pauli operator to be subtracted.
        """
        self.extend(rhs.copy(), -1.0)
        return self

    def extend(
        self,
        other: PauliOp | PauliString,
        multiplier: complex | None = None,
        dedupe: bool = True,
    ) -> None:
        """Add another PauliOp to the current one
        by extending the internal summation with new terms.

        Note: this does a shallow copy of other's properties

        Args:
        ----
            other: The Pauli operator to be added.
            multiplier: The coefficient to be applied to the new terms.
            dedupe: Flag to remove duplicate terms.
        """  # noqa: D205, D210
        if self.dim != other.dim:
            raise ValueError("Pauli operators must have the same dimension.")

        if isinstance(other, PauliString):
            self.pauli_strings.append(other)
            self.coeffs = np.append(self.coeffs, multiplier or 1.0)
        elif isinstance(other, PauliOp):
            self.pauli_strings += other.pauli_strings
            self.coeffs = np.append(
                self.coeffs, other.coeffs * multiplier if multiplier else other.coeffs
            )
        else:
            raise ValueError("Invalid type for the operand.")

        if dedupe:
            dedupe_strings: dict[PauliString, complex] = {}
            for c, p_str in zip(self.coeffs, self.pauli_strings):
                dedupe_strings[p_str] = dedupe_strings.get(p_str, 0) + c
            self.pauli_strings = list(dedupe_strings.keys())
            self.coeffs = np.array(list(dedupe_strings.values()), dtype=np.complex128)



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Module to validate correctness and consistency of C++ and python implementations."""



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Test Qiskit interface."""

from typing import Callable

import numpy as np
import pytest

import fast_pauli as fp


@pytest.mark.parametrize("size", [1, 2, 3, 4, 5])
def test_qiskit_interface(
    size: int, pauli_strings_with_size: Callable, generate_random_complex: Callable
) -> None:
    """Make sure we can go from fast_pauli to qiskit and back."""
    paulis = pauli_strings_with_size(size)
    op = fp.PauliOp(generate_random_complex(len(paulis)), paulis)
    q_op = fp.to_qiskit(op)
    fp_op = fp.from_qiskit(q_op)
    np.testing.assert_allclose(op.coeffs, fp_op.coeffs)
    np.testing.assert_array_equal(op.pauli_strings_as_str, fp_op.pauli_strings_as_str)


@pytest.mark.parametrize("n_qubits", [1, 2, 3, 4])
def test_qiskit_interface_pauli_string(
    n_qubits: int, pauli_strings_with_size: Callable
) -> None:
    """Make sure we can go from fast_pauli to qiskit and back."""
    for ps in pauli_strings_with_size(n_qubits):
        pauli = fp.PauliString(ps)
        q_pauli = fp.to_qiskit(pauli)

        # We don't currently save the phase for a PauliString in fast_pauli
        # so we need to make sure it's set to 1. Note that Qiskit keeps track
        # of the *exponent* of the phase, so it should be 0 here.
        assert q_pauli.phase == 0

        assert pauli.n_qubits == q_pauli.num_qubits

        # Check that the dense representation matches
        ps_matrix = pauli.to_tensor()
        q_matrix = q_pauli.to_matrix()
        np.testing.assert_allclose(ps_matrix, q_matrix)



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################

"""Process benchmark data from test_qiskit_adv.py."""

import os
import sys
from typing import Any

import pandas as pd
import plotly.express as px

FILE_DIR = os.path.dirname(os.path.abspath(__file__))
OUTPUT_DIR = os.path.join(FILE_DIR, "../../docs/benchmark_results/figs")
os.makedirs(OUTPUT_DIR, exist_ok=True)


################################################################################
# Helper functions
################################################################################
def format_fix(fig: Any) -> None:
    """Add consistent formatting to a plotly figure."""
    fig.update_layout(
        yaxis_title="Time (s)",
        template="seaborn",
        font=dict(size=18),
    )
    fig.update_traces(marker=dict(size=12))
    fig.update_yaxes(exponentformat="power")


COLOR_MAP = {
    "qiskit": "#0530AD",  # IBM blue
    "fast_pauli": "#fb8501",  # orange
}


################################################################################
# Load data
################################################################################

# WARNING DANGEROUS CLI (just don't wanna deal with argparse or click)
if len(sys.argv) != 2:
    print(f"Usage: {sys.argv[0]} <csv_file>")
    sys.exit(1)

df = pd.read_csv(sys.argv[1])

# Rename columns to be more readable in final HTML plots
df.rename(
    columns={
        "param:n_qubits": "N<sub>qubits</sub>",
        "param:n_pauli_strings": "N<sub>pauli strings</sub>",
        "param:method": "Library",
        "param:n_states": "N<sub>states</sub>",
    },
    inplace=True,
)


def get_sorted_unique_n_pauli_strings(df: pd.DataFrame) -> list[int]:
    """Get the sorted unique number of pauli strings from a dataframe.

    Used to make sure that the subplots are ordering consistently.
    """
    if "N<sub>pauli strings</sub>" in df.columns:
        sorted_unique_n_pauli_strings: list[int] = (
            df["N<sub>pauli strings</sub>"].dropna().unique().tolist()
        )
        sorted_unique_n_pauli_strings.sort()
        return sorted_unique_n_pauli_strings
    else:
        rv: list[int] = []
        return rv


def get_sorted_unique_n_states(df: pd.DataFrame) -> list[int]:
    """Get the sorted unique number of states from a dataframe.

    Used to make sure that the subplots are ordering consistently.
    """
    if "N<sub>states</sub>" in df.columns:
        sorted_unique_n_states: list[int] = (
            df["N<sub>states</sub>"].dropna().unique().tolist()
        )
        sorted_unique_n_states.sort()
        return sorted_unique_n_states
    else:
        rv: list[int] = []
        return rv


################################################################################
# Pauli string apply
################################################################################
df_ps = df[df["name"].str.contains("test_pauli_string_apply")]

fig = px.scatter(
    df_ps,
    x="N<sub>qubits</sub>",
    y="mean",
    error_y="stddev",
    color="Library",
    color_discrete_map=COLOR_MAP,
    # facet_col="N<sub>pauli strings</sub>",
    log_y=True,
    # title="Pauli String Applied to a State Vector",
)
format_fix(fig)

fig.write_html(
    f"{OUTPUT_DIR}/qiskit_pauli_string_apply.html",
    full_html=False,
    include_plotlyjs="cdn",
)

################################################################################
# Pauli Op applied to a statevector
################################################################################
df_op_app = df[df["name"].str.contains("test_pauli_op_apply")]

fig = px.scatter(
    df_op_app,
    x="N<sub>qubits</sub>",
    y="mean",
    error_y="stddev",
    color="Library",
    color_discrete_map=COLOR_MAP,
    facet_col="N<sub>pauli strings</sub>",
    log_y=True,
    # Make sure the subplots are ordered correctly
    category_orders={
        "N<sub>pauli strings</sub>": get_sorted_unique_n_pauli_strings(df_op_app),
    },
    # title="Pauli Operator Applied to a State Vector",
)

format_fix(fig)
fig.write_html(
    f"{OUTPUT_DIR}/qiskit_pauli_op_apply.html",
    full_html=False,
    include_plotlyjs="cdn",
)

################################################################################
# Pauli Op Expectation Value Batch
################################################################################
df_op_exp_batch = df[df["name"].str.contains("test_pauli_op_expectation_value_batch")]

fig = px.scatter(
    df_op_exp_batch,
    x="N<sub>qubits</sub>",
    y="mean",
    error_y="stddev",
    color="Library",
    color_discrete_map=COLOR_MAP,
    log_y=True,
    facet_col="N<sub>states</sub>",
    category_orders={
        "N<sub>states</sub>": get_sorted_unique_n_states(df_op_exp_batch),
    },
    # title="Expectation Value of Pauli Operator for Batch of States",
)
format_fix(fig)
fig.write_html(
    f"{OUTPUT_DIR}/qiskit_pauli_op_expectation_value_batch.html",
    full_html=False,
    include_plotlyjs="cdn",
)



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Test pauli operator objects from c++ and python."""

from typing import Callable

import numpy as np
import pytest

import fast_pauli as fp
import fast_pauli.pypauli as pp
from fast_pauli.pypauli.helpers import naive_pauli_converter, naive_pauli_operator
from tests.conftest import resolve_parameter_repr


def test_naive_pauli_operator(paulis: dict) -> None:
    """Test naive pauli operator helper for dense matrix representation."""
    for p in ["I", "X", "Y", "Z"]:
        np.testing.assert_equal(naive_pauli_operator([2], [p]), 2 * paulis[p])

    np.testing.assert_equal(naive_pauli_operator([1], ["III"]), np.eye(8))

    np.testing.assert_equal(
        naive_pauli_operator([2, 2, 2, 2], ["I", "X", "Y", "Z"]),
        2 * np.sum([paulis["I"], paulis["X"], paulis["Y"], paulis["Z"]], axis=0),
    )

    np.testing.assert_equal(
        naive_pauli_operator([3j, 3j, 3j], ["XY", "ZI", "YZ"]),
        3j
        * np.sum(
            [
                np.kron(paulis["X"], paulis["Y"]),
                np.kron(paulis["Z"], paulis["I"]),
                np.kron(paulis["Y"], paulis["Z"]),
            ],
            axis=0,
        ),
    )


@pytest.mark.parametrize(
    "pauli_op,", [fp.PauliOp, pp.PauliOp], ids=resolve_parameter_repr
)
def test_operator_trivial(
    paulis: dict,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
) -> None:
    """Test PauliOp class on trivial inputs."""
    empty = (
        [pauli_op(), pauli_op([]), pauli_op([], [])]
        if isinstance(pauli_op, fp.PauliOp)
        else [pauli_op([], [])]
    )
    for empty_op in empty:
        assert empty_op.dim == 0
        assert empty_op.n_qubits == 0
        assert empty_op.n_pauli_strings == 0
        assert len(empty_op.to_tensor()) == 0
        assert len(empty_op.coeffs) == 0
        assert len(empty_op.pauli_strings) == 0

    for p in ["I", "X", "Y", "Z"]:
        po = pauli_op([1.0], [p])
        assert po.dim == 2
        assert po.n_qubits == 1
        assert po.n_pauli_strings == 1
        # assert po.pauli_strings == [p]
        np.testing.assert_allclose(po.to_tensor(), paulis[p], atol=1e-15)


@pytest.mark.parametrize(
    "pauli_op,", [fp.PauliOp, pp.PauliOp], ids=resolve_parameter_repr
)
def test_operator_basics(
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
) -> None:
    """Test PauliOp class in python land."""
    po = pauli_op([4j, 4j], ["III", "III"])
    assert po.dim == 8
    assert po.n_qubits == 3
    assert po.n_pauli_strings == 2
    assert po.pauli_strings_as_str == ["III", "III"]
    np.testing.assert_equal(po.coeffs, [4j, 4j])
    np.testing.assert_allclose(po.to_tensor(), 4j * 2 * np.eye(8), atol=1e-15)

    string_sets = [
        ["I", "X", "Y", "Z"],
        pauli_strings_with_size(2),
        pauli_strings_with_size(3),
        ["XYZXYZXYZ", "ZZZIIIXXX"],
    ]
    if isinstance(po, fp.PauliOp):
        string_sets += [
            [fp.PauliString("XYXYX"), fp.PauliString("YXYXY"), fp.PauliString("ZXZXZ")],
        ]

    for strings in string_sets:
        coeffs = generate_random_complex(len(strings))
        po = pauli_op(coeffs, strings)

        n_qubits = (
            len(strings[0]) if isinstance(strings[0], str) else strings[0].n_qubits
        )
        assert po.dim == 2**n_qubits
        assert po.n_qubits == n_qubits
        assert po.n_pauli_strings == len(strings)

        assert set(po.pauli_strings_as_str) == set([str(ps) for ps in strings])
        np.testing.assert_allclose(
            po.coeffs,
            coeffs,
            atol=1e-15,
        )

        np.testing.assert_allclose(
            po.to_tensor(),
            naive_pauli_operator(coeffs, [str(ps) for ps in strings]),
            atol=1e-15,
        )
        np.testing.assert_allclose(
            po.to_tensor(),
            pauli_op(list(reversed(coeffs)), list(reversed(strings))).to_tensor(),
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,", [fp.PauliOp, pp.PauliOp], ids=resolve_parameter_repr
)
def test_dense_representation(
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
) -> None:
    """Test dense representation of Pauli Operator against naive approach."""
    for strings in [
        pauli_strings_with_size(2),
        pauli_strings_with_size(3),
        pauli_strings_with_size(4),
        pauli_strings_with_size(7, limit=128),
        pauli_strings_with_size(10, limit=64),
    ]:
        coeffs = generate_random_complex(len(strings))
        po = pauli_op(coeffs, strings)

        np.testing.assert_allclose(
            po.to_tensor(),
            naive_pauli_operator(coeffs, strings),
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,", [fp.PauliOp, pp.PauliOp], ids=resolve_parameter_repr
)
def test_apply_1d(
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
) -> None:
    """Test Pauli Operator multiplication with vector."""
    np.testing.assert_allclose(
        pauli_op([1, 2], ["IXYZ", "ZYXI"]).apply(np.zeros(16)),
        np.zeros(16),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        pauli_op([0.5, 0.5], ["III", "III"]).apply(np.arange(8)),
        np.arange(8),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        pauli_op([1j], ["XYZ"]).apply(np.ones(8)),
        naive_pauli_operator([1j], ["XYZ"]).sum(1),
        atol=1e-15,
    )

    for strings in [
        pauli_strings_with_size(2),
        pauli_strings_with_size(3),
        pauli_strings_with_size(4),
        pauli_strings_with_size(4),
        pauli_strings_with_size(7, limit=128),
        pauli_strings_with_size(8, limit=200)[:100],
        pauli_strings_with_size(8, limit=200)[100:],
        pauli_strings_with_size(10, limit=64),
    ]:
        coeffs = generate_random_complex(len(strings))
        po = pauli_op(coeffs, strings)

        psi = generate_random_complex(po.dim)
        np.testing.assert_allclose(
            po.apply(psi),
            naive_pauli_operator(coeffs, strings).dot(psi),
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,", [fp.PauliOp, pp.PauliOp], ids=resolve_parameter_repr
)
def test_apply_batch(
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
) -> None:
    """Test Pauli Operator multiplication with a batch of vectors."""
    np.testing.assert_allclose(
        pauli_op([1, 2], ["IXYZ", "ZYXI"]).apply(np.zeros((16, 16))),
        np.zeros((16, 16)),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        pauli_op([0.5, 0.5], ["III", "III"]).apply(np.arange(8 * 8).reshape(8, 8)),
        np.arange(8 * 8).reshape(8, 8),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        pauli_op([1j], ["XYZ"]).apply(np.eye(8)),
        naive_pauli_operator([1j], ["XYZ"]),
        atol=1e-15,
    )

    for strings in [
        pauli_strings_with_size(2),
        pauli_strings_with_size(3),
        pauli_strings_with_size(4),
        pauli_strings_with_size(7, limit=128),
        pauli_strings_with_size(8, limit=200)[:100],
        pauli_strings_with_size(8, limit=200)[100:],
        pauli_strings_with_size(10, limit=64),
    ]:
        n_states = int(100 * generate_random_complex(1)[0].real)
        coeffs = generate_random_complex(len(strings))
        po = pauli_op(coeffs, strings)

        psis = generate_random_complex(po.dim, n_states)
        np.testing.assert_allclose(
            po.apply(psis),
            naive_pauli_operator(coeffs, strings) @ psis,
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,", [fp.PauliOp, pp.PauliOp], ids=resolve_parameter_repr
)
def test_expectation_value(
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
) -> None:
    """Test Pauli Operator expected value."""
    np.testing.assert_allclose(
        pauli_op([1, 2], ["IXYZ", "ZYXI"]).expectation_value(np.zeros(16)),
        0,
        atol=1e-15,
    )
    np.testing.assert_allclose(
        pauli_op([1, 2], ["IXYZ", "ZYXI"]).expectation_value(np.zeros((16, 16))),
        np.zeros(16),
        atol=1e-15,
    )

    np.testing.assert_allclose(
        pauli_op([1, 1], ["III", "III"]).expectation_value(np.arange(8)),
        2 * np.square(np.arange(8)).sum(),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        pauli_op([1, 1], ["III", "III"]).expectation_value(
            np.arange(8 * 8).reshape(8, 8)
        ),
        2 * np.square(np.arange(8 * 8)).reshape(8, 8).sum(0),
        atol=1e-15,
    )

    for strings in [
        pauli_strings_with_size(2),
        pauli_strings_with_size(3),
        pauli_strings_with_size(4),
        pauli_strings_with_size(7, limit=128),
        pauli_strings_with_size(8, limit=200)[:100],
        pauli_strings_with_size(8, limit=200)[100:],
        pauli_strings_with_size(10, limit=64),
    ]:
        coeffs = generate_random_complex(len(strings))
        po = pauli_op(coeffs, strings)

        psi = generate_random_complex(po.dim)
        np.testing.assert_allclose(
            po.expectation_value(psi),
            naive_pauli_operator(coeffs, strings).dot(psi).dot(psi.conj()),
            atol=1e-15,
        )

        n_states = int(100 * generate_random_complex(1)[0].real)
        psis = generate_random_complex(n_states, po.dim)
        # compute <psi_t|A|psi_t>
        expected = np.einsum(
            "ti,ij,tj->t", psis.conj(), naive_pauli_operator(coeffs, strings), psis
        )
        np.testing.assert_allclose(
            po.expectation_value(psis.T.copy()),
            expected,
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,",
    [fp.PauliOp, pp.PauliOp],
    ids=resolve_parameter_repr,
)
def test_scaling_by_factor(
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
) -> None:
    """Test Pauli Operator scaling by a factor."""
    ixyz_op = pauli_op([1, 1, 1, 1], ["I", "X", "Y", "Z"])
    ixyz_expected = naive_pauli_operator([1, 1, 1, 1], ["I", "X", "Y", "Z"])

    np.testing.assert_allclose(
        (2 * ixyz_op * 2).to_tensor(),
        4 * ixyz_expected,
        atol=1e-15,
    )

    ixyz_op *= 4
    ixyz_op *= 0.25
    np.testing.assert_allclose(
        ixyz_op.to_tensor(),
        ixyz_expected,
        atol=1e-15,
    )

    ixyz_op.scale(-1)
    np.testing.assert_allclose(
        ixyz_op.to_tensor(),
        -ixyz_expected,
        atol=1e-15,
    )

    ixyz_op.scale(0)
    np.testing.assert_allclose(
        ixyz_op.to_tensor(),
        np.zeros((2, 2)),
        atol=1e-15,
    )

    n_runs = 5
    for strings in [
        pauli_strings_with_size(2),
        pauli_strings_with_size(3),
        pauli_strings_with_size(4),
        pauli_strings_with_size(7, limit=128),
        pauli_strings_with_size(8, limit=200)[:100],
        pauli_strings_with_size(8, limit=200)[100:],
    ]:
        for _ in range(n_runs):
            coeffs = generate_random_complex(len(strings))
            p_op = pauli_op(coeffs, strings)
            expected_op = naive_pauli_operator(coeffs, strings)

            factor = generate_random_complex(1)[0]
            np.testing.assert_allclose(
                (factor * p_op).to_tensor(),
                factor * expected_op,
                atol=1e-15,
            )
            np.testing.assert_allclose(
                (p_op * factor).to_tensor(),
                factor * expected_op,
                atol=1e-15,
            )

            factor = generate_random_complex(1)[0]
            p_op *= factor
            np.testing.assert_allclose(
                p_op.to_tensor(),
                factor * expected_op,
                atol=1e-15,
            )

            p_op.scale(1.0 / factor)
            np.testing.assert_allclose(
                p_op.to_tensor(),
                expected_op,
                atol=1e-15,
            )

            factors = generate_random_complex(len(strings))
            p_op.scale(factors)
            expected_op = naive_pauli_operator(factors * coeffs, strings)
            np.testing.assert_allclose(
                p_op.to_tensor(),
                expected_op,
                atol=1e-15,
            )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,pauli_string,",
    [(fp.PauliOp, fp.PauliString), (pp.PauliOp, pp.PauliString)],
    ids=resolve_parameter_repr,
)
def test_multiplication_with_string(
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test Pauli Operator multiplication with another Pauli Operator."""
    ixyz_op = pauli_op([1, 1, 1, 1], ["I", "X", "Y", "Z"])
    ixyz_expected = naive_pauli_operator([1, 1, 1, 1], ["I", "X", "Y", "Z"])

    np.testing.assert_allclose(
        (ixyz_op @ pauli_string("I")).to_tensor(),
        ixyz_expected,
        atol=1e-15,
    )
    np.testing.assert_allclose(
        (ixyz_op @ pauli_string("I") @ pauli_string("I")).to_tensor(),
        ixyz_expected,
        atol=1e-15,
    )
    np.testing.assert_allclose(
        (pauli_string("I") @ ixyz_op @ pauli_string("I")).to_tensor(),
        ixyz_expected,
        atol=1e-15,
    )

    for strings in [
        pauli_strings_with_size(2),
        pauli_strings_with_size(3),
        pauli_strings_with_size(4),
        pauli_strings_with_size(7, limit=128),
        pauli_strings_with_size(8, limit=200)[:100],
        pauli_strings_with_size(8, limit=200)[100:],
    ]:
        choose_string = int(generate_random_complex(1)[0].real * 100) % len(strings)
        coeffs = generate_random_complex(len(strings))
        p_str, coeffs = strings.pop(choose_string), np.delete(coeffs, choose_string)
        p_op = pauli_op(coeffs, strings)

        np.testing.assert_allclose(
            (p_op @ pauli_string(p_str)).to_tensor(),
            naive_pauli_operator(coeffs, strings) @ naive_pauli_converter(p_str),
            atol=1e-15,
        )
        np.testing.assert_allclose(
            (pauli_string(p_str) @ p_op).to_tensor(),
            naive_pauli_converter(p_str) @ naive_pauli_operator(coeffs, strings),
            atol=1e-15,
        )

    for strings, n_mult in [
        (pauli_strings_with_size(3), 25),
        (pauli_strings_with_size(4), 20),
        (pauli_strings_with_size(7, limit=64), 10),
        (pauli_strings_with_size(10, limit=32), 5),
    ]:
        coeffs = generate_random_complex(len(strings))
        p_op = pauli_op(coeffs, strings)
        expected_op = naive_pauli_operator(coeffs, strings)

        for _ in range(n_mult):
            choose_string = int(generate_random_complex(1)[0].real * 100) % len(strings)
            p_str = strings[choose_string]
            p_op = p_op @ pauli_string(p_str)
            expected_op @= naive_pauli_converter(p_str)

            np.testing.assert_allclose(
                p_op.to_tensor(),
                expected_op,
                atol=1e-15,
            )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,", [fp.PauliOp, pp.PauliOp], ids=resolve_parameter_repr
)
def test_multiplication_with_pauli_op(
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
) -> None:
    """Test Pauli Operator multiplication with another Pauli Operator."""
    ixyz_op = pauli_op([1, 1, 1, 1], ["I", "X", "Y", "Z"])
    ixyz_expected = naive_pauli_operator([1, 1, 1, 1], ["I", "X", "Y", "Z"])

    np.testing.assert_allclose(
        (pauli_op([1], ["I"]) @ pauli_op([1], ["I"])).to_tensor(),
        np.eye(2),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        (ixyz_op @ pauli_op([1], ["I"])).to_tensor(),
        ixyz_expected,
        atol=1e-15,
    )
    np.testing.assert_allclose(
        (pauli_op([1], ["I"]) @ ixyz_op @ pauli_op([1], ["I"])).to_tensor(),
        ixyz_expected,
        atol=1e-15,
    )

    rng = np.random.RandomState(42)
    for strings, variations in [
        (pauli_strings_with_size(2), 15),
        (pauli_strings_with_size(3), 10),
        (pauli_strings_with_size(4), 7),
        (pauli_strings_with_size(7, limit=128), 5),
        (pauli_strings_with_size(8, limit=200)[100:], 3),
        (pauli_strings_with_size(10, limit=32), 2),
    ]:
        for _ in range(variations):
            l_size = rng.choice(len(strings)) or 1
            r_size = rng.choice(len(strings)) or 1
            l_strings = rng.choice(strings, l_size)
            r_strings = rng.choice(strings, r_size)
            l_coeffs = generate_random_complex(l_size)
            r_coeffs = generate_random_complex(r_size)
            l_op = pauli_op(l_coeffs, l_strings)
            r_op = pauli_op(r_coeffs, r_strings)

            np.testing.assert_allclose(
                (l_op @ r_op).to_tensor(),
                naive_pauli_operator(l_coeffs, l_strings)
                @ naive_pauli_operator(r_coeffs, r_strings),
                atol=1e-15,
            )
            np.testing.assert_allclose(
                (r_op @ r_op).to_tensor(),
                naive_pauli_operator(r_coeffs, r_strings)
                @ naive_pauli_operator(r_coeffs, r_strings),
                atol=1e-15,
            )

        np.testing.assert_allclose(
            (l_op @ r_op @ l_op @ r_op).to_tensor(),
            naive_pauli_operator(l_coeffs, l_strings)
            @ naive_pauli_operator(r_coeffs, r_strings)
            @ naive_pauli_operator(l_coeffs, l_strings)
            @ naive_pauli_operator(r_coeffs, r_strings),
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,pauli_string,",
    [(fp.PauliOp, fp.PauliString), (pp.PauliOp, pp.PauliString)],
    ids=resolve_parameter_repr,
)
def test_pauli_string_conversion(
    pauli_strings_with_size: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test that addition of two pauli strings is resulting in expected pauli_op."""
    rng = np.random.RandomState(42)
    all_3qubit_strings = pauli_strings_with_size(3)
    all_4qubit_strings = pauli_strings_with_size(3)
    rng.shuffle(all_3qubit_strings)
    rng.shuffle(all_4qubit_strings)

    for string_set in [all_3qubit_strings, all_4qubit_strings]:
        while len(string_set) >= 2:
            l_str = string_set.pop()
            r_str = string_set.pop()

            np.testing.assert_allclose(
                (pauli_string(l_str) + pauli_string(r_str)).to_tensor(),
                naive_pauli_converter(l_str) + naive_pauli_converter(r_str),
                atol=1e-15,
            )
            np.testing.assert_allclose(
                (pauli_string(l_str) + pauli_string(r_str)).to_tensor(),
                pauli_op([1, 1], [l_str, r_str]).to_tensor(),
                atol=1e-15,
            )

            np.testing.assert_allclose(
                (pauli_string(l_str) - pauli_string(r_str)).to_tensor(),
                naive_pauli_converter(l_str) - naive_pauli_converter(r_str),
                atol=1e-15,
            )
            np.testing.assert_allclose(
                (pauli_string(l_str) - pauli_string(r_str)).to_tensor(),
                pauli_op([1, -1], [l_str, r_str]).to_tensor(),
                atol=1e-15,
            )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,pauli_string,",
    [(fp.PauliOp, fp.PauliString), (pp.PauliOp, pp.PauliString)],
    ids=resolve_parameter_repr,
)
def test_add_sub_with_string(
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test Pauli Operator addition with Pauli String."""
    ixyz_op = pauli_op([1, 1, 1, 1], ["I", "X", "Y", "Z"])

    np.testing.assert_allclose(
        (ixyz_op + pauli_string("X")).to_tensor(),
        naive_pauli_operator([1, 2, 1, 1], ["I", "X", "Y", "Z"]),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        (pauli_string("I") + ixyz_op + pauli_string("I")).to_tensor(),  # type: ignore[operator]
        naive_pauli_operator([3, 1, 1, 1], ["I", "X", "Y", "Z"]),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        (pauli_string("Z") + ixyz_op - pauli_string("Z")).to_tensor(),  # type: ignore[operator]
        naive_pauli_operator([1, 1, 1, 1], ["I", "X", "Y", "Z"]),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        (
            pauli_op([1j, 1j, 1j], ["X", "Y", "Z"])
            + pauli_op([1j, 1j, 1j], ["Z", "Y", "X"])
        ).to_tensor(),
        naive_pauli_operator([2j, 2j, 2j], ["Y", "X", "Z"]),
        atol=1e-15,
    )

    ixyz_op += pauli_string("Z")
    np.testing.assert_allclose(
        (ixyz_op).to_tensor(),
        naive_pauli_operator([1, 1, 1, 2], ["I", "X", "Y", "Z"]),
        atol=1e-15,
    )
    ixyz_op += pauli_string("X")
    np.testing.assert_allclose(
        (ixyz_op).to_tensor(),
        naive_pauli_operator([1, 2, 1, 2], ["I", "X", "Y", "Z"]),
        atol=1e-15,
    )
    ixyz_op -= pauli_string("Z")
    ixyz_op -= pauli_string("X")
    np.testing.assert_allclose(
        (ixyz_op).to_tensor(),
        naive_pauli_operator([1, 1, 1, 1], ["I", "X", "Y", "Z"]),
        atol=1e-15,
    )

    n_passes = 3
    for strings in [
        pauli_strings_with_size(2),
        pauli_strings_with_size(3),
        pauli_strings_with_size(4),
        pauli_strings_with_size(7, limit=128),
        pauli_strings_with_size(8, limit=200)[100:],
        pauli_strings_with_size(10, limit=32),
    ]:
        for _ in range(n_passes):
            choose_string = int(generate_random_complex(1)[0].real * len(strings))
            coeffs = generate_random_complex(len(strings))
            p_str, coeffs = strings.pop(choose_string), np.delete(coeffs, choose_string)
            p_op = pauli_op(coeffs, strings)
            expected_dense_op = naive_pauli_operator(coeffs, strings)
            p_str_dense = naive_pauli_converter(p_str)

            np.testing.assert_allclose(
                (p_op + pauli_string(p_str)).to_tensor(),
                expected_dense_op + p_str_dense,
                atol=1e-15,
            )
            np.testing.assert_allclose(
                (pauli_string(p_str) + p_op).to_tensor(),  # type: ignore[operator]
                p_str_dense + expected_dense_op,
                atol=1e-15,
            )

            np.testing.assert_allclose(
                (pauli_string(p_str) - p_op).to_tensor(),  # type: ignore[operator]
                p_str_dense - expected_dense_op,
                atol=1e-15,
            )
            np.testing.assert_allclose(
                (pauli_string(p_str) + p_op - pauli_string(p_str)).to_tensor(),  # type: ignore[operator]
                expected_dense_op,
                atol=1e-15,
            )

        choose_string = choose_string % len(strings)
        p_str, coeffs = strings.pop(choose_string), np.delete(coeffs, choose_string)
        p_op = pauli_op(coeffs, strings)
        expected = naive_pauli_operator(coeffs, strings)
        p_str_dense = naive_pauli_converter(p_str)

        p_op += pauli_string(p_str)
        expected += p_str_dense
        np.testing.assert_allclose(
            p_op.to_tensor(),
            expected,
            atol=1e-15,
        )
        p_op -= pauli_string(p_str)
        expected -= p_str_dense
        np.testing.assert_allclose(
            p_op.to_tensor(),
            expected,
            atol=1e-15,
        )

        p_op = pauli_op(coeffs, strings)
        c = generate_random_complex(1)[0]
        p_op.extend(pauli_string(p_str), c, dedupe=False)
        expected += c * p_str_dense
        np.testing.assert_allclose(
            p_op.to_tensor(),
            expected,
            atol=1e-15,
        )
        c = generate_random_complex(1)[0]
        p_op.extend(pauli_string(p_str), c, dedupe=True)
        expected += c * p_str_dense
        np.testing.assert_allclose(
            p_op.to_tensor(),
            expected,
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,", [fp.PauliOp, pp.PauliOp], ids=resolve_parameter_repr
)
def test_add_sub_with_pauli_op(
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
) -> None:
    """Test Pauli Operator addition with another Pauli Operator."""
    np.testing.assert_allclose(
        (pauli_op([1], ["II"]) + pauli_op([2], ["II"])).to_tensor(),
        3 * np.eye(4),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        (pauli_op([1j], ["X"]) + pauli_op([1j], ["Y"])).to_tensor(),
        naive_pauli_operator([1j, 1j], ["X", "Y"]),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        (pauli_op([1j], ["X"]) - pauli_op([1j], ["Y"])).to_tensor(),
        naive_pauli_operator([1j, -1j], ["X", "Y"]),
        atol=1e-15,
    )

    ixyz_op = pauli_op([1, 1, 1, 1], ["I", "X", "Y", "Z"])
    ixyz_expected = naive_pauli_operator([1, 1, 1, 1], ["I", "X", "Y", "Z"])

    np.testing.assert_allclose(
        (ixyz_op + ixyz_op - ixyz_op + ixyz_op).to_tensor(),
        2 * ixyz_expected,
        atol=1e-15,
    )
    np.testing.assert_allclose(
        (ixyz_op - ixyz_op).to_tensor(),
        naive_pauli_operator([0], ["X"]),
        atol=1e-15,
    )

    ixyz_op += pauli_op([1, 1, 1, 1], ["I", "X", "Y", "Z"])
    ixyz_op += ixyz_op
    np.testing.assert_allclose(
        ixyz_op.to_tensor(),
        4 * ixyz_expected,
        atol=1e-15,
    )

    ixyz_op -= pauli_op([4, 4], ["X", "Z"])
    np.testing.assert_allclose(
        ixyz_op.to_tensor(),
        naive_pauli_operator([4, 4], ["Y", "I"]),
        atol=1e-15,
    )
    ixyz_op -= pauli_op([3, 3], ["Y", "I"])
    np.testing.assert_allclose(
        ixyz_op.to_tensor(),
        naive_pauli_operator([1, 1], ["Y", "I"]),
        atol=1e-15,
    )

    rng = np.random.RandomState(42)
    for strings, variations in [
        (pauli_strings_with_size(2), 15),
        (pauli_strings_with_size(3), 10),
        (pauli_strings_with_size(4), 7),
        (pauli_strings_with_size(7, limit=128), 5),
        (pauli_strings_with_size(8, limit=200)[100:], 3),
        (pauli_strings_with_size(10, limit=32), 2),
    ]:
        for _ in range(variations):
            l_size = rng.choice(len(strings)) or 1
            r_size = rng.choice(len(strings)) or 1
            l_strings = rng.choice(strings, l_size)
            r_strings = rng.choice(strings, r_size)
            l_coeffs = generate_random_complex(l_size)
            r_coeffs = generate_random_complex(r_size)
            l_op = pauli_op(l_coeffs, l_strings)
            r_op = pauli_op(r_coeffs, r_strings)
            l_op_expected_dense = naive_pauli_operator(l_coeffs, l_strings)
            r_op_expected_dense = naive_pauli_operator(r_coeffs, r_strings)

            np.testing.assert_allclose(
                (l_op + r_op).to_tensor(),
                l_op_expected_dense + r_op_expected_dense,
                atol=1e-15,
            )
            np.testing.assert_allclose(
                (r_op + r_op).to_tensor(),
                2 * r_op_expected_dense,
                atol=1e-15,
            )
            np.testing.assert_allclose(
                (l_op - r_op).to_tensor(),
                l_op_expected_dense - r_op_expected_dense,
                atol=1e-15,
            )

        np.testing.assert_allclose(
            ((l_op + r_op) + (l_op - r_op)).to_tensor(),
            l_op_expected_dense
            + r_op_expected_dense
            + l_op_expected_dense
            - r_op_expected_dense,
            atol=1e-13,
        )

        l_op += r_op
        np.testing.assert_allclose(
            (l_op).to_tensor(),
            l_op_expected_dense + r_op_expected_dense,
            atol=1e-13,
        )
        l_op -= r_op
        np.testing.assert_allclose(
            (l_op).to_tensor(),
            l_op_expected_dense,
            atol=1e-13,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_op,pauli_string,",
    [(fp.PauliOp, fp.PauliString), (pp.PauliOp, pp.PauliString)],
    ids=resolve_parameter_repr,
)
def test_exceptions(
    pauli_op: type[fp.PauliOp] | type[pp.PauliOp],
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test that exceptions are raised and propagated correctly."""
    if isinstance(pauli_op, fp.PauliOp):
        with np.testing.assert_raises(ValueError):
            pauli_op(["ABC"])
        with np.testing.assert_raises(ValueError):
            pauli_op(["XI", "XY", "XZ", "III"])
    else:
        with np.testing.assert_raises(ValueError):
            pauli_op([1], ["ABC"])
        with np.testing.assert_raises(ValueError):
            pauli_op([1, 1, 1, 1], ["XI", "XY", "XZ", "III"])
    with np.testing.assert_raises(ValueError):
        pauli_op([1, 1], ["X", "YZ"])
    with np.testing.assert_raises(ValueError):
        pauli_op([1, 2, 3], ["I", "X", "Y", "Z"])

    with np.testing.assert_raises(ValueError):
        pauli_op([-1, 1], ["IX", "YZ"]) @ pauli_op([1], ["III"])
    with np.testing.assert_raises(ValueError):
        pauli_op([-1, 1], ["IX", "YZ"]) @ pauli_string("X")
    with np.testing.assert_raises(ValueError):
        pauli_op([-1, 1], ["IX", "YZ"]) + pauli_op([1], ["X"])
    with np.testing.assert_raises(ValueError):
        p_op = pauli_op([-1, 1], ["IX", "YZ"])
        p_op += pauli_op([1], ["Z"])
    with np.testing.assert_raises(ValueError):
        p_op = pauli_op([-1, 1], ["IX", "YZ"])
        p_op -= pauli_op([1], ["ZZZ"])
    with np.testing.assert_raises(ValueError):
        pauli_op([-1, 1], ["IX", "YZ"]) + pauli_string("XXXX")
    with np.testing.assert_raises(ValueError):
        p_op = pauli_op([-1, 1], ["IX", "YZ"])
        p_op += pauli_string("Z")
    with np.testing.assert_raises(ValueError):
        p_op = pauli_op([-1, 1], ["IX", "YZ"])
        p_op -= pauli_string("ZZZ")
    with np.testing.assert_raises(ValueError):
        pauli_op([-1, 1], ["IX", "YZ"]) - pauli_op([1], ["X"])
    with np.testing.assert_raises(ValueError):
        pauli_op([-1, 1], ["IX", "YZ"]) - pauli_string("XXXX")

    with np.testing.assert_raises(ValueError):
        pauli_op([-1, 1], ["IX", "YZ"]).extend(pauli_string("X"), 1j, dedupe=True)
    with np.testing.assert_raises(ValueError):
        pauli_op([-1, 1], ["IX", "YZ"]).extend(pauli_string("XYZ"), 1j, dedupe=False)
    with np.testing.assert_raises(ValueError):
        pauli_op([-1, 1], ["IX", "YZ"]).extend(pauli_op([1j], ["XXX"]))

    with np.testing.assert_raises(ValueError):
        pauli_op([1, 1, 1], ["X", "Y", "Z"]).apply(np.ones(3))
    with np.testing.assert_raises(ValueError):
        pauli_op([1, 1, 1], ["X", "Y", "Z"]).apply(np.eye(4))

    with np.testing.assert_raises(ValueError):
        pauli_op([1, 1], ["XYZ", "ZYX"]).expectation_value(np.ones(32))
    with np.testing.assert_raises(ValueError):
        pauli_op([1, 1], ["XYZ", "ZYX"]).expectation_value(np.eye(16))


if __name__ == "__main__":
    pytest.main()



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Test pauli objects from c++ against python implementations."""

import itertools as it

import numpy as np
import pytest

import fast_pauli as fp
from tests.conftest import resolve_parameter_repr


@pytest.mark.parametrize("pauli,", [(fp.Pauli)], ids=resolve_parameter_repr)
def test_basics(paulis: dict, pauli: type[fp.Pauli]) -> None:
    """Test pauli wrapper in python land."""
    np.testing.assert_array_equal(
        pauli().to_tensor(),
        paulis["I"],
    )

    for i in [0, 1, 2, 3]:
        pcpp = pauli(code=i)
        np.testing.assert_array_equal(
            pcpp.to_tensor(),
            paulis[i],
        )

    for c in ["I", "X", "Y", "Z"]:
        pcpp = pauli(symbol=c)
        np.testing.assert_allclose(
            pcpp.to_tensor(),
            paulis[c],
            atol=1e-15,
        )
        np.testing.assert_string_equal(str(pcpp), c)


@pytest.mark.parametrize("pauli,", [(fp.Pauli)], ids=resolve_parameter_repr)
def test_multiply(paulis: dict, pauli: type[fp.Pauli]) -> None:
    """Test custom __mul__ in c++ wrapper."""
    for p1, p2 in it.product("IXYZ", repeat=2):
        c, pcpp = pauli(p1) @ pauli(p2)
        np.testing.assert_allclose(
            c * np.array(pcpp.to_tensor()),
            paulis[p1] @ paulis[p2],
            atol=1e-15,
        )


@pytest.mark.parametrize("pauli,", [(fp.Pauli)], ids=resolve_parameter_repr)
def test_exceptions(pauli: type[fp.Pauli]) -> None:
    """Test that exceptions from c++ are raised and propagated correctly."""
    with np.testing.assert_raises(TypeError):
        pauli("II")
    with np.testing.assert_raises(ValueError):
        pauli("A")
    with np.testing.assert_raises(ValueError):
        pauli(-1)
    with np.testing.assert_raises(ValueError):
        pauli(5)


if __name__ == "__main__":
    pytest.main()



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Test pauli c++ objects against python implementations."""

import itertools as it
from typing import Callable

import numpy as np
import pytest

import fast_pauli as fp
import fast_pauli.pypauli as pp
from tests.conftest import (
    QUBITS_TO_BENCHMARK,
    SAMPLE_STRINGS_LIMIT,
    resolve_parameter_repr,
)

N_STATES_TO_BENCHMARK = [16, 128, 1024]


@pytest.fixture
def prepared_paulis(
    pauli_strings_with_size: Callable,
    pauli_class: type[fp.PauliString] | type[pp.PauliString],
    qubits: int,
) -> list:
    """Fixture to provide initialized Pauli strings for testing."""
    return list(
        map(
            lambda s: pauli_class(s),
            pauli_strings_with_size(qubits, limit=SAMPLE_STRINGS_LIMIT),
        )
    )


def benchmark_dense_conversion(paulis: list) -> None:
    """Benchmark dense conversion."""
    for p in paulis:
        dense_repr = p.to_tensor()  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_class,",
    it.chain(
        [(q, fp.PauliString) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliString) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_dense_conversion_n_qubits(
    benchmark: Callable,
    prepared_paulis: list,
    pauli_class: type[fp.PauliString] | type[pp.PauliString],
    qubits: int,
) -> None:
    """Benchmark dense conversion.

    Parametrized test case to run the benchmark across
    all Pauli strings of given length for given PauliString class.
    """
    benchmark(benchmark_dense_conversion, paulis=prepared_paulis)


def benchmark_apply(paulis: list, states: list) -> None:
    """Benchmark apply method."""
    for p, psi in zip(paulis, states):
        result = p.apply(psi)  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_class,",
    it.chain(
        [(q, fp.PauliString) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliString) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_apply_n_qubits(
    benchmark: Callable,
    prepared_paulis: list,
    generate_random_complex: Callable,
    pauli_class: type[fp.PauliString] | type[pp.PauliString],
    qubits: int,
) -> None:
    """Benchmark PauliString multiplication with provided state vector.

    Parametrized test case to run the benchmark across
    all Pauli strings of given length for given PauliString class.
    """
    n_dims = 1 << qubits
    prepared_states = [
        generate_random_complex(n_dims) for _ in range(len(prepared_paulis))
    ]

    benchmark(benchmark_apply, paulis=prepared_paulis, states=prepared_states)


@pytest.mark.parametrize(
    "qubits,states,pauli_class,",
    it.chain(
        [
            (q, n, fp.PauliString)
            for q in QUBITS_TO_BENCHMARK
            for n in N_STATES_TO_BENCHMARK
        ],
        [
            (q, n, pp.PauliString)
            for q in QUBITS_TO_BENCHMARK
            for n in N_STATES_TO_BENCHMARK
        ],
    ),
    ids=resolve_parameter_repr,
)
def test_apply_batch_n_qubits_n_states(
    benchmark: Callable,
    prepared_paulis: list,
    generate_random_complex: Callable,
    pauli_class: type[fp.PauliString] | type[pp.PauliString],
    qubits: int,
    states: int,
) -> None:
    """Benchmark PauliString multiplication with provided set of state vectors.

    Parametrized test case to run the benchmark across
    all Pauli strings of given length for given PauliString class.
    """
    n_dims = 1 << qubits
    prepared_states = [
        generate_random_complex(n_dims, states) for _ in range(len(prepared_paulis))
    ]

    benchmark(benchmark_apply, paulis=prepared_paulis, states=prepared_states)


def benchmark_expectation_value(paulis: list, states: list) -> None:
    """Benchmark expectation_value method."""
    for p, psi in zip(paulis, states):
        result = p.expectation_value(psi)  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_class,",
    it.chain(
        [(q, fp.PauliString) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliString) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_expectation_value_n_qubits(
    benchmark: Callable,
    prepared_paulis: list,
    generate_random_complex: Callable,
    pauli_class: type[fp.PauliString] | type[pp.PauliString],
    qubits: int,
) -> None:
    """Benchmark PauliString expectation_value with provided state vector."""
    n_dims = 1 << qubits
    prepared_states = [
        generate_random_complex(n_dims) for _ in range(len(prepared_paulis))
    ]

    benchmark(
        benchmark_expectation_value, paulis=prepared_paulis, states=prepared_states
    )


@pytest.mark.parametrize(
    "qubits,states,pauli_class,",
    it.chain(
        [
            (q, n, fp.PauliString)
            for q in QUBITS_TO_BENCHMARK
            for n in N_STATES_TO_BENCHMARK
        ],
        [
            (q, n, pp.PauliString)
            for q in QUBITS_TO_BENCHMARK
            for n in N_STATES_TO_BENCHMARK
        ],
    ),
    ids=resolve_parameter_repr,
)
def test_expectation_value_batch_n_qubits_n_states(
    benchmark: Callable,
    prepared_paulis: list,
    generate_random_complex: Callable,
    pauli_class: type[fp.PauliString] | type[pp.PauliString],
    qubits: int,
    states: int,
) -> None:
    """Benchmark PauliString expectation_value with provided set of state vectors."""
    n_dims = 1 << qubits
    prepared_states = [
        generate_random_complex(n_dims, states) for _ in range(len(prepared_paulis))
    ]

    benchmark(
        benchmark_expectation_value, paulis=prepared_paulis, states=prepared_states
    )


def benchmark_matmul(left_paulis: list, right_paulis: list) -> None:
    """Benchmark PauliString multiplication."""
    for lp, rp in zip(left_paulis, right_paulis):
        phase, p_str = lp @ rp  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_class,",
    it.chain(
        [(q, fp.PauliString) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliString) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_multiplication_n_qubits(
    benchmark: Callable,
    prepared_paulis: list,
    pauli_class: type[fp.PauliString] | type[pp.PauliString],
    qubits: int,
) -> None:
    """Benchmark matrix multiplication of PauliStrings.

    Parametrized test case to run the benchmark across
    all Pauli strings of given length for given PauliString class.
    """
    left_paulis, right_paulis = np.array_split(prepared_paulis, 2)

    benchmark(benchmark_matmul, left_paulis=left_paulis, right_paulis=right_paulis)


def benchmark_arithmetic(left_paulis: list, right_paulis: list) -> None:
    """Benchmark PauliString arithmetic."""
    for lp, rp in zip(left_paulis, right_paulis):
        p_op1 = lp + rp  # noqa: F841
        p_op2 = rp - lp  # noqa: F841


@pytest.mark.parametrize(
    "qubits,pauli_class,",
    it.chain(
        [(q, fp.PauliString) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.PauliString) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_arithmetic_n_qubits(
    benchmark: Callable,
    prepared_paulis: list,
    pauli_class: type[fp.PauliString] | type[pp.PauliString],
    qubits: int,
) -> None:
    """Benchmark addition and subtraction of PauliStrings.

    Parametrized test case to run the benchmark across
    all Pauli strings of given length for given PauliString class.
    """
    left_paulis, right_paulis = np.array_split(prepared_paulis, 2)

    benchmark(benchmark_arithmetic, left_paulis=left_paulis, right_paulis=right_paulis)


def benchmark_sparse_composer(paulis: list, composer: Callable) -> None:
    """Benchmark algorithm for sparse representation of pauli string."""
    for ps in paulis:
        cols, vals = composer(ps)  # noqa: F841


@pytest.mark.skip(
    reason="currently std::vector conversion on c++ side "
    "takes majority of time -> results are not representative"
)
@pytest.mark.parametrize(
    "qubits,composer_func,",
    it.chain(
        [(q, fp.helpers.pauli_string_sparse_repr) for q in QUBITS_TO_BENCHMARK],
        [(q, pp.pauli_string.compose_sparse_pauli) for q in QUBITS_TO_BENCHMARK],
    ),
    ids=resolve_parameter_repr,
)
def test_string_sparse_composer_n_qubits(
    benchmark: Callable,
    pauli_strings_with_size: Callable,
    composer_func: Callable,
    qubits: int,
) -> None:
    """Benchmark algorithm for sparse representation of pauli string.

    Parametrized test case to run the benchmark across
    all Pauli strings of given length for given PauliString class.
    """
    prepared_paulis: list = pauli_strings_with_size(qubits, limit=SAMPLE_STRINGS_LIMIT)
    if "pypauli" not in composer_func.__module__:  # check if it's c++ wrapper
        prepared_paulis = [[fp.Pauli(c) for c in s] for s in prepared_paulis]

    benchmark(benchmark_sparse_composer, paulis=prepared_paulis, composer=composer_func)


if __name__ == "__main__":
    pytest.main()



# noqa: D100
import os
import sys

if __name__ == "__main__":
    compiler = sys.argv[1]
    os_type = sys.argv[2]

    if "macos" not in os_type:
        os.system("sudo apt-get update")

        if "g++" in compiler and "clang" not in compiler:
            version = compiler.split("-")[1]
            os.system(f"sudo apt-get -y install gcc-{version} g++-{version}")
        elif "clang" in compiler:
            version = compiler.split("-")[1]
            print(f"Installing clang {compiler}")
            os.system("wget https://apt.llvm.org/llvm.sh")
            os.system("chmod +x llvm.sh")
            os.system(f"sudo ./llvm.sh {version}")
            os.system(f"sudo apt-get install -y libomp-{version}-dev")
        else:
            raise ValueError(f"Compiler {compiler} not supported")



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Test pauli objects from c++ against python implementations."""

import itertools as it
from typing import Callable

import numpy as np
import pytest

import fast_pauli as fp
import fast_pauli.pypauli as pp
from fast_pauli.pypauli.helpers import naive_pauli_converter
from tests.conftest import resolve_parameter_repr


def test_naive_pauli_converter(paulis: dict) -> None:
    """Test naive Pauli converter for dense matrix."""
    for p in ["I", "X", "Y", "Z"]:
        np.testing.assert_equal(naive_pauli_converter(p), paulis[p])

    np.testing.assert_equal(naive_pauli_converter("III"), np.eye(8))
    np.testing.assert_equal(naive_pauli_converter("XXX"), np.fliplr(np.eye(8)))

    np.testing.assert_equal(
        naive_pauli_converter("IZ"), np.kron(paulis["I"], paulis["Z"])
    )
    np.testing.assert_equal(
        naive_pauli_converter("IZ"),
        np.block([[paulis["Z"], np.zeros((2, 2))], [np.zeros((2, 2)), paulis["Z"]]]),
    )
    np.testing.assert_equal(
        naive_pauli_converter("IY"), np.kron(paulis["I"], paulis["Y"])
    )
    np.testing.assert_equal(
        naive_pauli_converter("IY"),
        np.block([[paulis["Y"], np.zeros((2, 2))], [np.zeros((2, 2)), paulis["Y"]]]),
    )
    np.testing.assert_equal(
        naive_pauli_converter("YX"), np.kron(paulis["Y"], paulis["X"])
    )
    np.testing.assert_equal(
        naive_pauli_converter("XYZ"),
        np.kron(paulis["X"], np.kron(paulis["Y"], paulis["Z"])),
    )


@pytest.mark.parametrize(
    "pauli_string,", [fp.PauliString, pp.PauliString], ids=resolve_parameter_repr
)
def test_string_trivial(
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test pauli string class on trivial inputs."""
    empty_paulis = (
        [pauli_string(), pauli_string([]), pauli_string("")]
        if isinstance(pauli_string, fp.PauliString)
        else [pauli_string("")]
    )
    for empty_ps in empty_paulis:
        assert empty_ps.weight == 0
        assert empty_ps.dim == 0
        assert empty_ps.n_qubits == 0
        assert str(empty_ps) == ""
        assert len(empty_ps.to_tensor()) == 0

    ps = pauli_string("III")
    assert ps.dim == 8
    assert ps.weight == 0
    np.testing.assert_equal(ps.to_tensor(), np.eye(8))


@pytest.mark.parametrize(
    "pauli_string,", [fp.PauliString, pp.PauliString], ids=resolve_parameter_repr
)
def test_string_basics(
    paulis: dict,
    pauli_strings_with_size: Callable,
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test pauli string class in python land."""
    for s in it.chain(
        ["I", "X", "Y", "Z"],
        [[fp.Pauli("I")], [fp.Pauli("X")], [fp.Pauli("Y")], [fp.Pauli("Z")]]
        if isinstance(pauli_string, fp.PauliString)
        else [],
    ):
        p = pauli_string(s)  # type: ignore[arg-type]
        assert p.weight == 1 or str(p) == "I"
        assert p.dim == 2
        assert p.n_qubits == 1
        if isinstance(s, str):
            assert str(p) == s
            np.testing.assert_allclose(
                p.to_tensor(),
                paulis[s],
                atol=1e-15,
            )

    for s in pauli_strings_with_size(3):
        p = pauli_string(s)  # type: ignore[arg-type]
        assert p.weight == len(s) - s.count("I")
        assert p.dim == 8
        assert p.n_qubits == 3
        assert str(p) == s
        np.testing.assert_allclose(
            p.to_tensor(),
            np.kron(paulis[s[0]], np.kron(paulis[s[1]], paulis[s[2]])),
            atol=1e-15,
        )

    assert pauli_string("XYIZXYZ").dim == 2**7
    assert pauli_string("XYIZXYZ").weight == 6
    assert pauli_string("XXIYYIZZ").dim == 2**8
    assert pauli_string("XXIYYIZZ").weight == 6
    assert pauli_string("ZIXIZYXXY").dim == 2**9
    assert pauli_string("ZIXIZYXXY").weight == 7


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_string,", [fp.PauliString, pp.PauliString], ids=resolve_parameter_repr
)
def test_dense_representation(
    paulis: dict,
    sample_pauli_strings: list,
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test dense representation of Pauli strings against naive approach."""
    ps = pauli_string("II")
    assert ps.weight == 0
    assert ps.dim == 4
    np.testing.assert_equal(ps.to_tensor(), np.eye(4))

    ps = pauli_string("IIII")
    assert ps.weight == 0
    assert ps.dim == 16
    np.testing.assert_equal(ps.to_tensor(), np.eye(16))

    ixyz = np.array(pauli_string("IXYZ").to_tensor())
    np.testing.assert_allclose(ixyz, naive_pauli_converter("IXYZ"), atol=1e-15)

    zyxi = np.array(pauli_string("ZYXI").to_tensor())
    np.testing.assert_allclose(zyxi, naive_pauli_converter("ZYXI"), atol=1e-15)

    assert np.abs(ixyz - zyxi).sum().sum() >= 1

    for s in sample_pauli_strings:
        ps = pauli_string(s)
        assert ps.weight == len(s) - s.count("I")
        assert ps.dim == 2 ** len(s)
        assert str(ps) == s
        np.testing.assert_allclose(
            ps.to_tensor(),
            naive_pauli_converter(s),
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_string,", [fp.PauliString, pp.PauliString], ids=resolve_parameter_repr
)
def test_apply_1d(
    sample_pauli_strings: list,
    generate_random_complex: Callable,
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test pauli string multiplication with 1d vector."""
    np.testing.assert_allclose(
        pauli_string("IXYZ").apply(np.zeros(16)),
        np.zeros(16),
        atol=1e-15,
    )

    np.testing.assert_allclose(
        pauli_string("III").apply(np.arange(8)),
        np.arange(8),
        atol=1e-15,
    )

    np.testing.assert_allclose(
        pauli_string("ZYX").apply(np.ones(8)),
        naive_pauli_converter("ZYX").sum(1),
        atol=1e-15,
    )

    for s in sample_pauli_strings:
        n_dim = 2 ** len(s)
        psi = generate_random_complex(n_dim)
        np.testing.assert_allclose(
            pauli_string(s).apply(psi),
            naive_pauli_converter(s).dot(psi),
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_string,", [fp.PauliString, pp.PauliString], ids=resolve_parameter_repr
)
def test_apply_batch(
    sample_pauli_strings: list,
    pauli_strings_with_size: Callable,
    generate_random_complex: Callable,
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test pauli string multiplication with 2d tensor."""
    np.testing.assert_allclose(
        pauli_string("IXYZ").apply(np.zeros((16, 16))),
        np.zeros((16, 16)),
        atol=1e-15,
    )

    np.testing.assert_allclose(
        pauli_string("III").apply(np.arange(8 * 8).reshape(8, 8)),
        np.arange(8 * 8).reshape(8, 8),
        atol=1e-15,
    )

    np.testing.assert_allclose(
        pauli_string("ZYX").apply(np.eye(8)),
        naive_pauli_converter("ZYX"),
        atol=1e-15,
    )

    for s in sample_pauli_strings:
        n_dim = 2 ** len(s)
        n_states = 42
        psis = generate_random_complex(n_dim, n_states)
        np.testing.assert_allclose(
            pauli_string(s).apply(psis),
            naive_pauli_converter(s) @ psis,
            atol=1e-15,
        )

    for s in pauli_strings_with_size(4):
        n_dim = 2 ** len(s)
        n_states = 7
        coeff = generate_random_complex(1)[0]
        psis = generate_random_complex(n_dim, n_states)

        np.testing.assert_allclose(
            pauli_string(s).apply(psis, coeff),
            coeff * naive_pauli_converter(s) @ psis,
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_string,", [(fp.PauliString), (pp.PauliString)], ids=resolve_parameter_repr
)
def test_expectation_value(
    sample_pauli_strings: list,
    generate_random_complex: Callable,
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test the expected value method."""
    np.testing.assert_allclose(
        pauli_string("IXYZ").expectation_value(np.zeros(16)),
        0,
        atol=1e-15,
    )
    np.testing.assert_allclose(
        pauli_string("IXYZ").expectation_value(np.zeros((16, 16))),
        np.zeros(16),
        atol=1e-15,
    )

    np.testing.assert_allclose(
        pauli_string("III").expectation_value(np.arange(8)),
        np.square(np.arange(8)).sum(),
        atol=1e-15,
    )
    np.testing.assert_allclose(
        pauli_string("III").expectation_value(np.arange(8 * 8).reshape(8, 8)),
        np.square(np.arange(8 * 8)).reshape(8, 8).sum(0),
        atol=1e-15,
    )

    for s in sample_pauli_strings:
        n_dim = 2 ** len(s)
        n_states = 21

        psi = generate_random_complex(n_dim)
        np.testing.assert_allclose(
            pauli_string(s).expectation_value(psi),
            naive_pauli_converter(s).dot(psi).dot(psi.conj()),
            atol=1e-15,
        )

        psis = generate_random_complex(n_states, n_dim)
        # compute <psi_t|P_i|psi_t>
        expected = np.einsum("ti,ij,tj->t", psis.conj(), naive_pauli_converter(s), psis)
        np.testing.assert_allclose(
            pauli_string(s).expectation_value(psis.T.copy()),
            expected,
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_string,", [fp.PauliString, pp.PauliString], ids=resolve_parameter_repr
)
def test_multiplication(
    sample_pauli_strings: list,
    pauli_strings_with_size: Callable,
    paulis: dict,
    pauli_string: type[fp.PauliString] | type[pp.PauliString],
) -> None:
    """Test pauli string multiplication with 2d tensor."""
    sign, pauli = pauli_string("I") @ (pauli_string("I"))
    np.testing.assert_equal(sign, 1)
    np.testing.assert_string_equal(str(pauli), "I")
    np.testing.assert_allclose(
        pauli.to_tensor(),
        paulis["I"],
        atol=1e-15,
    )
    sign, pauli = pauli_string("IIII") @ (pauli_string("IIII"))
    np.testing.assert_equal(sign, 1)
    np.testing.assert_string_equal(str(pauli), "IIII")
    np.testing.assert_allclose(
        pauli.to_tensor(),
        naive_pauli_converter("IIII"),
        atol=1e-15,
    )
    sign, pauli = pauli_string("XYZ") @ (pauli_string("III"))
    np.testing.assert_equal(sign, 1)
    np.testing.assert_string_equal(str(pauli), "XYZ")
    np.testing.assert_allclose(
        pauli.to_tensor(),
        naive_pauli_converter("XYZ"),
        atol=1e-15,
    )
    sign, pauli = pauli_string("ZYX") @ (pauli_string("ZYX"))
    np.testing.assert_equal(sign, 1)
    np.testing.assert_string_equal(str(pauli), "III")
    np.testing.assert_allclose(
        pauli.to_tensor(),
        np.eye(8),
        atol=1e-15,
    )

    for full_str in it.chain(
        pauli_strings_with_size(2),
        pauli_strings_with_size(4),
        pauli_strings_with_size(8, limit=100),
        pauli_strings_with_size(12, limit=10),
    ):
        mid = len(full_str) // 2
        p_str1, p_str2 = full_str[:mid], full_str[mid:]

        phase, pauli = pauli_string(p_str1) @ (pauli_string(p_str2))
        np.testing.assert_allclose(
            phase * np.array(pauli.to_tensor()),
            naive_pauli_converter(p_str1) @ naive_pauli_converter(p_str2),
            atol=1e-15,
        )
        phase, pauli = pauli_string(p_str2) @ (pauli_string(p_str1))
        np.testing.assert_allclose(
            phase * np.array(pauli.to_tensor()),
            naive_pauli_converter(p_str2) @ naive_pauli_converter(p_str1),
            atol=1e-15,
        )


@pytest.mark.consistency
@pytest.mark.parametrize(
    "pauli_string,", [(fp.PauliString), (pp.PauliString)], ids=resolve_parameter_repr
)
def test_exceptions(pauli_string: type[fp.PauliString] | type[pp.PauliString]) -> None:
    """Test that exceptions are raised and propagated correctly."""
    with np.testing.assert_raises(ValueError):
        pauli_string("ABC")
    with np.testing.assert_raises(ValueError):
        pauli_string("xyz")

    if isinstance(pauli_string, fp.PauliString):
        with np.testing.assert_raises(TypeError):
            pauli_string([1, 2])
        with np.testing.assert_raises(AttributeError):
            pauli_string("XYZ").dim = 99
        with np.testing.assert_raises(AttributeError):
            pauli_string("XYZ").weight = 99

    with np.testing.assert_raises(ValueError):
        pauli_string("II").apply(np.array([0.1, 0.2, 0.3]))
    with np.testing.assert_raises(ValueError):
        pauli_string("XYZ").apply(np.eye(4))


@pytest.mark.consistency
def test_sparse_composers(paulis: dict, pauli_strings_with_size: Callable) -> None:
    """Test consistency for sparse pauli composers."""
    np.testing.assert_equal(
        pp.pauli_string.compose_sparse_diag_pauli("IZI"),
        pp.pauli_string.compose_sparse_pauli("IZI")[1],
    )

    np.testing.assert_equal(
        pp.pauli_string.compose_sparse_diag_pauli("ZIZI"),
        pp.pauli_string.compose_sparse_pauli("ZIZI")[1],
    )

    np.testing.assert_equal(
        pp.pauli_string.compose_sparse_diag_pauli("ZZZIII"),
        pp.pauli_string.compose_sparse_pauli("ZZZIII")[1],
    )

    for s in ["I", "X", "Y", "Z"] + pauli_strings_with_size(3):
        py_cols, py_vals = pp.pauli_string.compose_sparse_pauli(s)
        cpp_cols, cpp_vals = fp.helpers.pauli_string_sparse_repr(
            [fp.Pauli(c) for c in s]
        )
        np.testing.assert_allclose(py_cols, cpp_cols, atol=1e-15)
        np.testing.assert_allclose(py_vals, cpp_vals, atol=1e-15)


if __name__ == "__main__":
    pytest.main()



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################

"""Process benchmark data."""

import itertools
import os

import pandas as pd
import plotly.express as px

FILE_DIR = os.path.dirname(os.path.abspath(__file__))
OUTPUT_DIR = os.path.join(FILE_DIR, "../../docs/benchmark_results/figs")

df = pd.read_csv("benchmark_results.csv")

functions_to_save = [
    # "test_dense_conversion_n_qubits",
    "test_apply_n_qubits",
    "test_apply_batch_n_qubits_n_states",
    "test_expectation_value_n_qubits",
    "test_expectation_value_batch_n_qubits_n_states",
    "test_multiplication_n_qubits",
    "test_arithmetic_n_qubits",
    # "test_string_sparse_composer_n_qubits",
]

# Extract what Class is being tested
# WARNING: This is brittle and assumes the class name is either
# "PauliString" or "PauliOp"
df.loc[:, "class"] = [
    "PauliString" if "PauliString" in name else "PauliOp" for name in df["name"]
]


for f, c in itertools.product(functions_to_save, ["PauliString", "PauliOp"]):
    # General processing

    # Simplify the name
    df_f = df[df["name"].str.contains(f) & (df["class"] == c)]
    if df_f.empty:
        continue

    df_f.loc[:, "name"] = df_f["name"].str.split("::").str[-1]

    # Extract C++/Python impl details
    df_f.loc[:, "impl"] = ["cpp" if "cpp" in name else "py" for name in df_f["name"]]

    # Use n_states as facet_col if it exists
    if df_f["param:states"].isna().any():
        fig = px.scatter(
            df_f,
            x="param:qubits",
            y="mean",
            error_y="stddev",
            color="impl",
            log_y=True,
        )

    else:
        fig = px.scatter(
            df_f,
            x="param:qubits",
            y="mean",
            error_y="stddev",
            color="impl",
            facet_col="param:states",
            log_y=True,
        )

    fig.update_layout(
        # title=f"Benchmark: {f} for class {c}",
        xaxis_title="Number of Qubits",
        yaxis_title="Time (s)",
        template="plotly_white",
        font=dict(size=18),
    )
    fig.update_traces(marker=dict(size=12))
    # fig.show()

    os.makedirs(OUTPUT_DIR, exist_ok=True)
    fig.write_html(
        f"{OUTPUT_DIR}/{f}_{c}.html",
        full_html=False,
        include_plotlyjs="cdn",
    )



#############################################################################
# This code is part of Fast Pauli.
#
# (C) Copyright Qognitive Inc 2024.
#
# This code is licensed under the BSD 2-Clause License. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#############################################################################


"""Test SummedPauliOp Python and C++ implementations."""

import numpy as np
import pytest

import fast_pauli as fp
import fast_pauli.pypauli as pp
from tests.conftest import resolve_parameter_repr


@pytest.mark.parametrize("n_qubits,n_operators", [(2, 2), (3, 2), (4, 3)])
def test_ctors(n_qubits: int, n_operators: int) -> None:
    """Test the constructor of SummedPauliOp."""
    # Test with PauliStrings
    pauli_strings = fp.helpers.calculate_pauli_strings_max_weight(n_qubits, 2)
    coeffs_2d = np.random.rand(len(pauli_strings), n_operators).astype(np.complex128)
    op = fp.SummedPauliOp(pauli_strings, coeffs_2d)
    assert op.dim == 2**n_qubits
    assert op.n_operators == n_operators
    assert op.n_pauli_strings == len(pauli_strings)

    # Test with list of strings
    pauli_strings_str = [str(s) for s in pauli_strings]
    op = fp.SummedPauliOp(pauli_strings_str, coeffs_2d)
    assert op.dim == 2**n_qubits
    assert op.n_operators == n_operators
    assert op.n_pauli_strings == len(pauli_strings)


@pytest.mark.parametrize(
    "summed_pauli_op", [fp.SummedPauliOp], ids=resolve_parameter_repr
)
@pytest.mark.parametrize(
    "n_states,n_operators,n_qubits",
    [(s, o, q) for s in [1, 10, 1000] for o in [1, 10, 100] for q in [1, 2, 6]],
)
def test_apply(
    summed_pauli_op: type[fp.SummedPauliOp],
    n_states: int,
    n_operators: int,
    n_qubits: int,
) -> None:
    """Test applying the summed pauli operator method."""
    pauli_strings = fp.helpers.calculate_pauli_strings_max_weight(n_qubits, 2)
    n_strings = len(pauli_strings)

    coeffs_2d = np.random.rand(n_strings, n_operators).astype(np.complex128)
    psi = np.random.rand(2**n_qubits, n_states).astype(np.complex128)

    # For a single state, test that a 1D array works
    if n_states == 1:
        psi = psi[:, 0]

    op = summed_pauli_op(pauli_strings, coeffs_2d)

    # The new_states we want to check
    new_states = op.apply(psi)

    # Trusted new_states
    new_states_naive = np.zeros((2**n_qubits, n_states), dtype=np.complex128)
    # For a single state, test that a 1D array works
    if n_states == 1:
        new_states_naive = new_states_naive[:, 0]

    for k in range(n_operators):
        A_k = fp.PauliOp(coeffs_2d[:, k].copy(), pauli_strings)
        new_states_naive += A_k.apply(psi)

    # Check
    np.testing.assert_allclose(new_states, new_states_naive, atol=1e-13)


@pytest.mark.parametrize(
    "summed_pauli_op", [fp.SummedPauliOp], ids=resolve_parameter_repr
)
@pytest.mark.parametrize(
    "n_states,n_operators,n_qubits",
    [(s, o, q) for s in [1, 10, 1000] for o in [1, 10, 100] for q in [1, 2, 6]],
)
def test_apply_weighted(
    summed_pauli_op: type[fp.SummedPauliOp],
    n_states: int,
    n_operators: int,
    n_qubits: int,
) -> None:
    """Test applying the summed pauli operator method."""
    pauli_strings = fp.helpers.calculate_pauli_strings_max_weight(n_qubits, 2)
    n_strings = len(pauli_strings)

    coeffs_2d = np.random.rand(n_strings, n_operators).astype(np.complex128)
    psi = np.random.rand(2**n_qubits, n_states).astype(np.complex128)
    data_weights = np.random.rand(n_operators, n_states).astype(np.float64)

    # For a single state, test that a 1D array works
    if n_states == 1:
        psi = psi[:, 0]
        data_weights = data_weights[:, 0]

    op = summed_pauli_op(pauli_strings, coeffs_2d)

    # The new_states we want to check
    new_states = op.apply_weighted(psi, data_weights)

    # Trusted new_states
    new_states_naive = np.zeros((2**n_qubits, n_states), dtype=np.complex128)
    # For a single state, test that a 1D array works
    if n_states == 1:
        new_states_naive = new_states_naive[:, 0]

    for k in range(n_operators):
        A_k = fp.PauliOp(coeffs_2d[:, k].copy(), pauli_strings)
        new_states_naive += A_k.apply(psi) * data_weights[k]

    # Check
    np.testing.assert_allclose(new_states, new_states_naive, atol=1e-13)


@pytest.mark.parametrize(
    "summed_pauli_op", [fp.SummedPauliOp], ids=resolve_parameter_repr
)
@pytest.mark.parametrize(
    "n_states,n_operators,n_qubits",
    [(s, o, q) for s in [1, 10, 1000] for o in [1, 10, 100] for q in [1, 2, 6]],
)
def test_expectation_values(
    summed_pauli_op: type[fp.SummedPauliOp],
    n_states: int,
    n_operators: int,
    n_qubits: int,
) -> None:
    """Test expectation value calculation."""
    # TODO This will break on python version
    pauli_strings = fp.helpers.calculate_pauli_strings_max_weight(n_qubits, 2)
    pauli_strings_str = [str(s) for s in pauli_strings]
    n_strings = len(pauli_strings)

    coeffs_2d = np.random.rand(n_strings, n_operators).astype(np.complex128)

    if n_states == 1:
        psi = np.random.rand(2**n_qubits).astype(np.complex128)
        op = summed_pauli_op(pauli_strings, coeffs_2d)
        # The expectation values we want to test
        expectation_vals = op.expectation_value(psi)
        # The "trusted" expectation_values
        expectation_vals_naive = np.zeros((n_operators), dtype=np.complex128)
        for k in range(n_operators):
            A_k = pp.helpers.naive_pauli_operator(coeffs_2d[:, k], pauli_strings_str)
            expectation_vals_naive[k] = np.einsum("i,ij,j->", psi.conj(), A_k, psi)

    else:
        psi = np.random.rand(2**n_qubits, n_states).astype(np.complex128)
        op = summed_pauli_op(pauli_strings, coeffs_2d)
        # The expectation values we want to test
        expectation_vals = op.expectation_value(psi)
        # The "trusted" expectation_values
        expectation_vals_naive = np.zeros((n_operators, n_states), dtype=np.complex128)

        for k in range(n_operators):
            A_k = pp.helpers.naive_pauli_operator(coeffs_2d[:, k], pauli_strings_str)
            expectation_vals_naive[k] = np.einsum("it,ij,jt->t", psi.conj(), A_k, psi)

    # Check
    np.testing.assert_allclose(expectation_vals, expectation_vals_naive, atol=1e-13)


@pytest.mark.parametrize(
    "summed_pauli_op", [fp.SummedPauliOp], ids=resolve_parameter_repr
)
@pytest.mark.parametrize(
    "n_operators,n_qubits",
    [(o, q) for o in [1, 10, 100] for q in [1, 2, 6]],
)
def test_to_tensor(
    summed_pauli_op: type[fp.SummedPauliOp],
    n_operators: int,
    n_qubits: int,
) -> None:
    """Test the dense representation of the SummedPauliOp."""
    # initialize SummedPauliOp
    pauli_strings = fp.helpers.calculate_pauli_strings_max_weight(n_qubits, 2)
    n_strings = len(pauli_strings)
    coeffs_2d = np.random.rand(n_strings, n_operators).astype(np.complex128)
    op = summed_pauli_op(pauli_strings, coeffs_2d)

    # get dense representation
    dense_op = op.to_tensor()

    # Check against doing it manually
    # Get dense representation of each PauliString
    ps_dense = np.array([ps.to_tensor() for ps in pauli_strings])

    summed_pauli_op_check = np.zeros(
        (n_operators, 2**n_qubits, 2**n_qubits), dtype=np.complex128
    )

    for k in range(n_operators):
        for j in range(n_strings):
            summed_pauli_op_check[k] += coeffs_2d[j, k] * ps_dense[j]

    np.testing.assert_allclose(dense_op, summed_pauli_op_check)


@pytest.mark.parametrize(
    "summed_pauli_op", [fp.SummedPauliOp], ids=resolve_parameter_repr
)
@pytest.mark.parametrize(
    "n_operators,n_qubits",
    [(o, q) for o in [1, 10] for q in [1, 2, 4, 6]],
)
def test_square(
    summed_pauli_op: type[fp.SummedPauliOp],
    n_operators: int,
    n_qubits: int,
) -> None:
    """Test squaring the SummedPauliOp."""
    pauli_strings = fp.helpers.calculate_pauli_strings_max_weight(n_qubits, 2)
    coeffs_2d = np.random.rand(len(pauli_strings), n_operators).astype(np.complex128)
    op = summed_pauli_op(pauli_strings, coeffs_2d)
    op2 = op.square()

    A_k = op.to_tensor()
    A_k2 = op2.to_tensor()

    np.testing.assert_allclose(A_k2, np.einsum("kab,kbc->kac", A_k, A_k))
